#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils import parse_time

from ..enums import AllowedTimeUnitsSAFER, FeatureDiscoveryMode
from ..helpers.feature_discovery import DatasetDefinition, FeatureDiscoverySetting, Relationship


class RelationshipsConfiguration(APIObject):
    """A Relationships configuration specifies a set of secondary datasets as well as
    the relationships among them. It is used to configure Feature Discovery for a project
    to generate features automatically from these datasets.

    Attributes
    ----------
    id : string
        Id of the created relationships configuration
    dataset_definitions: list
        Each element is a dataset_definitions for a dataset.
    relationships: list
        Each element is a relationship between two datasets
    feature_discovery_mode: str
        Mode of feature discovery. Supported values are 'default' and 'manual'
    feature_discovery_settings: list
        List of feature discovery settings used to customize the feature discovery process

    The `dataset_definitions` structure is

    identifier: string
        Alias of the dataset (used directly as part of the generated feature names)
    catalog_id: str, or None
        Identifier of the catalog item
    catalog_version_id: str
        Identifier of the catalog item version
    primary_temporal_key: string, optional
        Name of the column indicating time of record creation
    feature_list_id: string, optional
        Identifier of the feature list. This decides which columns in the dataset are
        used for feature generation
    snapshot_policy: str
        Policy to use  when creating a project or making predictions.
        Must be one of the following values:
        'specified': Use specific snapshot specified by catalogVersionId
        'latest': Use latest snapshot from the same catalog item
        'dynamic': Get data from the source (only applicable for JDBC datasets)
    feature_lists: list
        List of feature list info
    data_source: dict
        Data source info if the dataset is from data source
    data_sources: list
        List of Data source details for a JDBC datasets
    is_deleted: bool, optional
        Whether the dataset is deleted or not

    The `data source info` structured is

    data_store_id: str
        Id of the data store.
    data_store_name : str
         User-friendly name of the data store.
    url : str
        Url used to connect to the data store.
    dbtable : str
        Name of table from the data store.
    schema: str
        Schema definition of the table from the data store
    catalog: str
        Catalog name of the data source.

    The `feature list info` structure is

    id : str
        Id of the featurelist
    name : str
        Name of the featurelist
    features : list of str
        Names of all the Features in the featurelist
    dataset_id : str
        Project the featurelist belongs to
    creation_date : datetime.datetime
        When the featurelist was created
    user_created : bool
        Whether the featurelist was created by a user or by DataRobot automation
    created_by: str
        Name of user who created it
    description : str
        Description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    dataset_id: str
        Dataset which is associated with the feature list
    dataset_version_id: str or None
        Version of the dataset which is associated with feature list.
        Only relevant for Informative features

    The `relationships` schema is

    dataset1_identifier: str or None
        Identifier of the first dataset in this relationship.
        This is specified in the identifier field of dataset_definition structure.
        If None, then the relationship is with the primary dataset.
    dataset2_identifier: str
        Identifier of the second dataset in this relationship.
        This is specified in the identifier field of dataset_definition schema.
    dataset1_keys: list of str (max length: 10 min length: 1)
        Column(s) from the first dataset which are used to join to the second dataset
    dataset2_keys: list of str (max length: 10 min length: 1)
        Column(s) from the second dataset that are used to join to the first dataset
    time_unit: str, or None
        Time unit of the feature derivation window. Supported
        values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_start: int, or None
        How many time_units of each dataset's primary temporal key into the past relative
        to the datetimePartitionColumn the feature derivation window should begin.
        Will be a negative integer,
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_end: int, or None
        How many timeUnits of each dataset's record
        primary temporal key into the past relative to the datetimePartitionColumn the
        feature derivation window should end.  Will be a non-positive integer, if present.
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_time_unit: int or None
        Time unit of the feature derivation window. Supported values are
        MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
        If present, time-aware joins will be used.
        Only applicable when dataset1Identifier is not provided.
    feature_derivation_windows: list of dict, or None
        List of feature derivation windows settings. If present, time-aware joins will be used.
        Only allowed when feature_derivation_window_start,
        feature_derivation_window_end and feature_derivation_window_time_unit are not provided.
    prediction_point_rounding: int, or None
        Closest value of prediction_point_rounding_time_unit to round the prediction point
        into the past when applying the feature derivation window. Will be a positive integer,
        if present.Only applicable when dataset1_identifier is not provided.
    prediction_point_rounding_time_unit: str, or None
        time unit of the prediction point rounding. Supported values are
        MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
        Only applicable when dataset1_identifier is not provided.

    The `feature_derivation_windows` is a list of dictionary with schema:
        start: int
            How many time_units of each dataset's primary temporal key into the past relative
            to the datetimePartitionColumn the feature derivation window should begin.
        end: int
            How many timeUnits of each dataset's record
            primary temporal key into the past relative to the datetimePartitionColumn the
            feature derivation window should end.
        unit: string
            Time unit of the feature derivation window.
            One of ``datarobot.enums.AllowedTimeUnitsSAFER``.

    The `feature_discovery_settings` structure is:

    name: str
        Name of the feature discovery setting
    value: bool
        Value of the feature discovery setting

    To see the list of possible settings, create a RelationshipConfiguration without specifying
    settings and check its `feature_discovery_settings` attribute, which is a list of possible
    settings with their default values.
    """

    _path = "relationshipsConfigurations/"
    data_source_trafaret = t.Dict(
        {
            t.Key("data_store_name"): String,
            t.Key("data_store_id"): String,
            t.Key("url", optional=True): t.Or(String, t.Null),
            t.Key("dbtable", optional=True): t.Or(String, t.Null),
            t.Key("schema", optional=True): t.Or(String(allow_blank=True), t.Null),
            t.Key("catalog", optional=True): t.Or(String(allow_blank=True), t.Null),
            t.Key("data_source_id", optional=True): t.Or(String, t.Null),
        }
    ).ignore_extra("*")
    feature_list_info = t.Dict(
        {
            t.Key("id"): String,
            t.Key("features"): t.List(String),
            t.Key("name"): String,
            t.Key("description"): String,
            t.Key("dataset_id"): String,
            t.Key("user_created"): t.Bool,
            t.Key("creation_date"): parse_time,
            t.Key("created_by"): String,
            t.Key("dataset_version_id", optional=True): String,
        }
    ).ignore_extra("*")
    dataset_definitions_trafaret = t.Dict(
        {
            t.Key("identifier"): String(min_length=3, max_length=20),
            t.Key("catalog_version_id"): String,
            t.Key("catalog_id"): String,
            t.Key("primary_temporal_key", optional=True): t.Or(String, t.Null),
            t.Key("feature_list_id", optional=True): t.Or(String, t.Null),
            t.Key("snapshot_policy", optional=True, default="latest"): t.Enum(
                "latest", "specified", "dynamic"
            ),
            t.Key("feature_lists", optional=True): t.List(feature_list_info),
            t.Key("data_source", optional=True): t.Or(data_source_trafaret, t.Null),
            t.Key("data_sources", optional=True): t.Or(t.List(data_source_trafaret), t.Null),
            t.Key("is_deleted", optional=True): t.Or(t.Bool, t.Null),
        }
    ).ignore_extra("*")
    feature_derivation_window_trafaret = t.Dict(
        {
            t.Key("start"): Int(lt=0),
            t.Key("end"): Int(lte=0),
            t.Key("unit"): t.Enum(*AllowedTimeUnitsSAFER.ALL),
        }
    )

    relationships_trafaret = t.Dict(
        {
            t.Key("dataset1_identifier", optional=True): t.Or(String, t.Null),
            t.Key("dataset2_identifier"): String,
            t.Key("dataset1_keys"): t.List(String, min_length=1, max_length=10),
            t.Key("dataset2_keys"): t.List(String, min_length=1, max_length=10),
            t.Key("feature_derivation_window_start", optional=True): Int(lt=0),
            t.Key("feature_derivation_window_end", optional=True): Int(lte=0),
            t.Key("feature_derivation_window_time_unit", optional=True): t.Enum(
                *AllowedTimeUnitsSAFER.ALL
            ),
            t.Key("feature_derivation_windows", optional=True): t.List(
                feature_derivation_window_trafaret
            ),
            t.Key("prediction_point_rounding", optional=True): Int(gt=0, lte=30),
            t.Key("prediction_point_rounding_time_unit", optional=True): t.Enum(
                *AllowedTimeUnitsSAFER.ALL
            ),
        }
    ).ignore_extra("*")

    feature_discovery_setting_trafaret = t.Dict(
        {t.Key("name"): String, t.Key("value"): t.Bool}
    ).ignore_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("dataset_definitions"): t.List(dataset_definitions_trafaret, min_length=1),
            t.Key("relationships"): t.List(relationships_trafaret, min_length=1),
            t.Key("feature_discovery_mode", optional=True): t.Enum(*FeatureDiscoveryMode.ALL),
            t.Key("feature_discovery_settings", optional=True): t.List(
                feature_discovery_setting_trafaret
            ),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id,
        dataset_definitions=None,
        relationships=None,
        feature_discovery_mode=None,
        feature_discovery_settings=None,
    ):
        self.id = id
        self.dataset_definitions = dataset_definitions
        self.relationships = relationships
        self.feature_discovery_mode = feature_discovery_mode
        self.feature_discovery_settings = feature_discovery_settings

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @classmethod
    def create(cls, dataset_definitions, relationships, feature_discovery_settings=None):
        """Create a Relationships Configuration

        Parameters
        ----------
        dataset_definitions: list of dataset definitions
            Each element is a ``datarobot.helpers.feature_discovery.DatasetDefinition``
        relationships: list of relationships
            Each element is a ``datarobot.helpers.feature_discovery.Relationship``
        feature_discovery_settings : list of feature discovery settings, optional
            Each element is a dictionary or a
            ``datarobot.helpers.feature_discovery.FeatureDiscoverySetting``. If not provided,
            default settings will be used.

        Returns
        -------
        relationships_configuration: RelationshipsConfiguration
            Created relationships configuration

        Examples
        --------
        .. code-block:: python

            dataset_definition = dr.DatasetDefinition(
                identifier='profile',
                catalog_id='5fd06b4af24c641b68e4d88f',
                catalog_version_id='5fd06b4af24c641b68e4d88f'
            )
            relationship = dr.Relationship(
                dataset2_identifier='profile',
                dataset1_keys=['CustomerID'],
                dataset2_keys=['CustomerID'],
                feature_derivation_window_start=-14,
                feature_derivation_window_end=-1,
                feature_derivation_window_time_unit='DAY',
                prediction_point_rounding=1,
                prediction_point_rounding_time_unit='DAY'
            )
            dataset_definitions = [dataset_definition]
            relationships = [relationship]
            relationship_config = dr.RelationshipsConfiguration.create(
                dataset_definitions=dataset_definitions,
                relationships=relationships,
                feature_discovery_settings = [
                    {'name': 'enable_categorical_statistics', 'value': True},
                    {'name': 'enable_numeric_skewness', 'value': True},
                ]
            )
            >>> relationship_config.id
            '5c88a37770fc42a2fcc62759'
        """
        payload_dataset_definition = []
        for dd in dataset_definitions:
            if isinstance(dd, DatasetDefinition):
                payload_dataset_definition.append(dd.to_payload())
            else:
                payload_dataset_definition.append(dd)

        payload_relationships = []
        for rel in relationships:
            if isinstance(rel, Relationship):
                payload_relationships.append(rel.to_payload())
            else:
                payload_relationships.append(rel)

        payload_data = {
            "datasetDefinitions": payload_dataset_definition,
            "relationships": payload_relationships,
        }

        if feature_discovery_settings is not None:
            settings = []
            for item in feature_discovery_settings:
                if isinstance(item, FeatureDiscoverySetting):
                    settings.append(item.to_dict())
                else:
                    settings.append(item)
            payload_data["feature_discovery_settings"] = settings

        response = cls._client.post(cls._path, data=payload_data).json()
        return RelationshipsConfiguration.from_server_data(response)

    def get(self):
        """Retrieve the Relationships configuration for a given id

        Returns
        -------
        relationships_configuration: RelationshipsConfiguration
            The requested relationships configuration

        Raises
        ------
        ClientError
            Raised if an invalid relationships config id is provided.

        Examples
        --------
        .. code-block:: python

            relationships_config = dr.RelationshipsConfiguration(valid_config_id)
            result = relationships_config.get()
            >>> result.id
            '5c88a37770fc42a2fcc62759'
        """
        return self.from_location(f"{self._path}{self.id}/")

    def replace(self, dataset_definitions, relationships, feature_discovery_settings=None):
        """Update the Relationships Configuration which is not used in
        the feature discovery Project

        Parameters
        ----------
        dataset_definitions: list of dataset definition
            Each element is a ``datarobot.helpers.feature_discovery.DatasetDefinition``
        relationships: list of relationships
            Each element is a ``datarobot.helpers.feature_discovery.Relationship``
        feature_discovery_settings : list of feature discovery settings, optional
            Each element is a dictionary or a
            ``datarobot.helpers.feature_discovery.FeatureDiscoverySetting``. If not provided,
            default settings will be used.

        Returns
        -------
        relationships_configuration: RelationshipsConfiguration
            the updated relationships configuration
        """
        payload_dataset_definition = []
        for dd in dataset_definitions:
            if isinstance(dd, DatasetDefinition):
                payload_dataset_definition.append(dd.to_payload())
            else:
                payload_dataset_definition.append(dd)

        payload_relationships = []
        for rel in relationships:
            if isinstance(rel, Relationship):
                payload_relationships.append(rel.to_payload())
            else:
                payload_relationships.append(rel)

        payload_data = {
            "datasetDefinitions": payload_dataset_definition,
            "relationships": payload_relationships,
        }

        if feature_discovery_settings is not None:
            settings = []
            for item in feature_discovery_settings:
                if isinstance(item, FeatureDiscoverySetting):
                    settings.append(item.to_dict())
                else:
                    settings.append(item)
            payload_data["feature_discovery_settings"] = settings

        url = f"{self._path}{self.id}/"
        response = self._client.put(url, json=payload_data).json()
        return RelationshipsConfiguration.from_server_data(response)

    def delete(self):
        """
        Delete the Relationships configuration

        Raises
        ------
        ClientError
            Raised if an invalid relationships config id is provided.

        Examples
        --------
        .. code-block:: python

            # Deleting with a valid id
            relationships_config = dr.RelationshipsConfiguration(valid_config_id)
            status_code = relationships_config.delete()
            status_code
            >>> 204
            relationships_config.get()
            >>> ClientError: Relationships Configuration not found
        """
        self._client.delete(f"{self._path}{self.id}/")



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.chat import Chat
from datarobot.models.genai.llm import LLMDefinition
from datarobot.models.genai.llm_blueprint import (
    LLMBlueprint,
    vector_database_settings_trafaret,
    VectorDatabaseSettings,
)
from datarobot.models.genai.playground import Playground
from datarobot.models.genai.vector_database import VectorDatabase
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


def get_entity_id(
    entity: Union[Playground, Chat, ChatPrompt, LLMBlueprint, LLMDefinition, VectorDatabase, str]
) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be entity ID or the entity.

    Returns
    -------
    id : str
        Entity ID
    """
    if isinstance(entity, str):
        return entity

    return entity.id


result_metadata_trafaret = t.Dict(
    {
        t.Key("cost", optional=True): t.Or(t.Float, t.Null),
        t.Key("output_token_count"): t.Int,
        t.Key("input_token_count"): t.Int,
        t.Key("total_token_count"): t.Int,
        t.Key("estimated_docs_token_count"): t.Int,
        t.Key("latency_milliseconds"): t.Int,
        t.Key("error_message", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


confidence_scores_trafaret = t.Dict(
    {
        t.Key("rouge"): t.Float,
        t.Key("meteor"): t.Float,
        t.Key("bleu"): t.Float,
    }
).ignore_extra("*")


citation_trafaret = t.Dict(
    {
        t.Key("text"): t.String,
        t.Key("source", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


chat_prompt_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("text"): t.String,
        t.Key("llm_blueprint_id"): t.String,
        t.Key("llm_id"): t.String,
        t.Key("llm_settings", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("vector_database_id", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_settings", optional=True): t.Or(
            vector_database_settings_trafaret, t.Null
        ),
        t.Key("result_metadata", optional=True): t.Or(result_metadata_trafaret, t.Null),
        t.Key("result_text", optional=True): t.Or(t.String, t.Null),
        t.Key("confidence_scores", optional=True): t.Or(confidence_scores_trafaret, t.Null),
        t.Key("citations"): t.List(citation_trafaret),
        t.Key("execution_status"): t.String,
        t.Key("chat_id", optional=True): t.Or(t.String, t.Null),
        t.Key("chat_context_id", optional=True): t.Or(t.String, t.Null),
        t.Key("chat_prompt_ids_included_in_history", optional=True): t.Or(t.List(t.String), t.Null),
    }
).ignore_extra("*")


class ResultMetadata(APIObject):
    """
    Metadata for the result of a chat prompt submission.

    Attributes
    ----------
    output_token_count : int
        The number of tokens in the output.
    input_token_count : int
        The number of tokens in the input. This includes the chat history and documents
        retrieved from a vector database, if any.
    total_token_count : int
        The total number of tokens processed.
    estimated_docs_token_count : int
        The estimated number of tokens from the documents retrieved from a vector database, if any.
    latency_milliseconds : int
        The latency of the chat prompt submission in milliseconds.
    error_message : str or None, optional
        The error message from the LLM response.
    cost : float or None, optional
        The cost of the chat prompt submission.
    """

    _converter = result_metadata_trafaret

    def __init__(
        self,
        output_token_count: int,
        input_token_count: int,
        total_token_count: int,
        estimated_docs_token_count: int,
        latency_milliseconds: int,
        error_message: Optional[str] = None,
        cost: Optional[float] = None,
    ):
        self.output_token_count = output_token_count
        self.input_token_count = input_token_count
        self.total_token_count = total_token_count
        self.estimated_docs_token_count = estimated_docs_token_count
        self.latency_milliseconds = latency_milliseconds
        self.error_message = error_message
        self.cost = cost

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(output_token_count={self.output_token_count}, "
            f"input_token_count={self.input_token_count}, "
            f"total_token_count={self.total_token_count}, "
            f"estimated_docs_token_count={self.estimated_docs_token_count}, "
            f"latency_milliseconds={self.latency_milliseconds})"
        )


class ConfidenceScores(APIObject):
    """
    Confidence scores for a chat prompt submission that uses a vector database.

    Attributes
    ----------
    rouge : float
        The rouge score
    meteor : float
        The meteor score
    bleu : float
        The bleu score
    """

    _converter = confidence_scores_trafaret

    def __init__(
        self,
        rouge: float,
        meteor: float,
        bleu: float,
    ):
        self.rouge = rouge
        self.meteor = meteor
        self.bleu = bleu

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(rouge={self.rouge}, "
            f"meteor={self.meteor}, bleu={self.bleu})"
        )


class Citation(APIObject):
    """
    Citation for documents retrieved from a vector database.

    Attributes
    ----------
    text : str
        The text retrieved from a vector database.
    source : str or None, optional
        The source of the retrieved text.
    """

    _converter = citation_trafaret

    def __init__(
        self,
        text: str,
        source: Optional[str] = None,
    ):
        self.text = text
        self.source = source

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text={self.text}, source={self.source})"


class ChatPrompt(APIObject):
    """
    Metadata for a DataRobot GenAI chat prompt.

    Attributes
    ----------
    id : str
        Chat prompt ID.
    text : str
        The prompt text.
    llm_blueprint_id : str
        ID of the LLM blueprint associated with the chat prompt.
    llm_id : str
        ID of the LLM type. This must be one of the IDs returned by `LLMDefinition.list`
        for this user.
    llm_settings : dict or None
        The LLM settings for the LLM blueprint. The specific keys allowed and the
        constraints on the values are defined in the response from `LLMDefinition.list`,
        but this typically has dict fields. Either:
        - system_prompt - The system prompt that influences the LLM responses.
        - max_completion_length - The maximum number of tokens in the completion.
        - temperature - Controls the variability in the LLM response.
        - top_p - Sets whether the model considers next tokens with top_p probability mass.
        or
        - system_prompt - The system prompt that influences the LLM responses.
        - validation_id - The ID of the external model LLM validation.
        - external_llm_context_size - The external LLM's context size, in tokens,
        for external model-based LLM blueprints.
    creation_date : str
        The date the chat prompt was created.
    creation_user_id : str
        ID of the creating user.
    vector_database_id : str or None
        ID of the vector database associated with the LLM blueprint, if any.
    vector_database_settings : VectorDatabaseSettings or None
        The settings for the vector database associated with the LLM blueprint, if any.
    result_metadata : ResultMetadata or None
        Metadata for the result of the chat prompt submission.
    result_text: str or None
        The result text from the chat prompt submission.
    confidence_scores: ConfidenceScores or None
        The confidence scores if there is a vector database associated with the chat prompt.
    citations: list[Citation]
        List of citations from text retrieved from the vector database, if any.
    execution_status: str
        The execution status of the chat prompt.
    chat_id: Optional[str]
        ID of the chat associated with the chat prompt.
    chat_context_id: Optional[str]
        The ID of the chat context for the chat prompt.
    chat_prompt_ids_included_in_history: Optional[list[str]]
        The IDs of the chat prompts included in the chat history for this chat prompt.
    """

    _path = "api/v2/genai/chatPrompts"

    _converter = chat_prompt_trafaret

    def __init__(
        self,
        id: str,
        text: str,
        llm_blueprint_id: str,
        llm_id: str,
        creation_date: str,
        creation_user_id: str,
        citations: List[Dict[str, Any]],
        execution_status: str,
        llm_settings: Optional[Dict[str, Any]] = None,
        vector_database_id: Optional[str] = None,
        vector_database_settings: Optional[Dict[str, Any]] = None,
        result_metadata: Optional[Dict[str, Any]] = None,
        result_text: Optional[str] = None,
        confidence_scores: Optional[Dict[str, float]] = None,
        chat_id: Optional[str] = None,
        chat_context_id: Optional[str] = None,
        chat_prompt_ids_included_in_history: Optional[list[str]] = None,
    ):
        self.id = id
        self.text = text
        self.llm_blueprint_id = llm_blueprint_id
        self.llm_id = llm_id
        self.llm_settings = llm_settings
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.citations = [Citation.from_server_data(citation) for citation in citations]
        self.execution_status = execution_status
        self.vector_database_id = vector_database_id
        self.vector_database_settings = (
            VectorDatabaseSettings.from_server_data(vector_database_settings)
            if vector_database_settings
            else None
        )
        self.result_metadata = (
            ResultMetadata.from_server_data(result_metadata) if result_metadata else None
        )
        self.result_text = result_text
        self.confidence_scores = (
            ConfidenceScores.from_server_data(confidence_scores) if confidence_scores else None
        )
        self.chat_id = chat_id
        self.chat_context_id = chat_context_id
        self.chat_prompt_ids_included_in_history = chat_prompt_ids_included_in_history

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id}, "
            f"execution_status={self.execution_status}, text={self.text[:1000]})"
        )

    @classmethod
    def create(
        cls,
        text: str,
        llm_blueprint: Optional[Union[LLMBlueprint, str]] = None,
        chat: Optional[Union[Chat, str]] = None,
        llm: Optional[Union[LLMDefinition, str]] = None,
        llm_settings: Optional[Dict[str, Optional[Union[bool, int, float, str]]]] = None,
        vector_database: Optional[Union[VectorDatabase, str]] = None,
        vector_database_settings: Optional[VectorDatabaseSettings] = None,
        wait_for_completion: bool = False,
    ) -> ChatPrompt:
        """
        Create a new ChatPrompt. This submits the prompt text to the LLM. Either `llm_blueprint`
        or `chat` is required.

        Parameters
        ----------
        text : str
            The prompt text.
        llm_blueprint : LLMBlueprint or str or None, optional
            The LLM blueprint associated with the created chat prompt, either `LLMBlueprint` or
            LLM blueprint ID.
        chat : Chat or str or None, optional
            The chat associated with the created chat prompt, either `Chat` or chat ID.
        llm : LLMDefinition, str, or None, optional
            LLM to use for the chat prompt. Accepts `LLMDefinition` or LLM ID.
        llm_settings: dict or None
            LLM settings to use for the chat prompt. The specific keys allowed and the
            constraints on the values are defined in the response from `LLMDefinition.list`
            but this typically has dict fields:
            - system_prompt - The system prompt that tells the LLM how to behave.
            - max_completion_length - The maximum number of token in the completion.
            - temperature - controls the variability in the LLM response.
            - top_p - the model considers next tokens with top_p probability mass
            or
            - system_prompt - The system prompt that tells the LLM how to behave.
            - validation_id - The ID of the custom model LLM validation
            for custom model LLM blueprints.
        vector_database: VectorDatabase, str, or None, optional
            The vector database to use with this chat prompt submission.
            Accepts `VectorDatabase` or vector database ID.
        vector_database_settings: VectorDatabaseSettings or None, optional
            Settings for the vector database, if any.
        wait_for_completion : bool
            If set to True code will wait for the chat prompt job to complete before
            returning the result (up to 10 minutes, raising timeout error after that).
            Otherwise, you can check current status by using `ChatPrompt.get` with returned ID.

        Returns
        -------
        chat_prompt : ChatPrompt
            The created chat prompt.
        """
        payload = {
            "llm_blueprint_id": get_entity_id(llm_blueprint) if llm_blueprint else None,
            "chat_id": get_entity_id(chat) if chat else None,
            "text": text,
            "llm_id": get_entity_id(llm) if llm else None,
            "llm_settings": llm_settings,
            "vector_database_id": get_entity_id(vector_database) if vector_database else None,
            "vector_database_settings": vector_database_settings.to_dict()
            if vector_database_settings
            else None,
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        if wait_for_completion:
            location = wait_for_async_resolution(cls._client, r_data.headers["Location"])
            return cls.from_location(location)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, chat_prompt: Union[ChatPrompt, str]) -> ChatPrompt:
        """
        Retrieve a single chat prompt.

        Parameters
        ----------
        chat_prompt : ChatPrompt or str
            The chat prompt you want to retrieve, either `ChatPrompt` or chat prompt ID.

        Returns
        -------
        chat_prompt : ChatPrompt
            The requested chat prompt.
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(chat_prompt)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        llm_blueprint: Optional[Union[LLMBlueprint, str]] = None,
        playground: Optional[Union[Playground, str]] = None,
        chat: Optional[Union[Chat, str]] = None,
    ) -> List[ChatPrompt]:
        """
        List all chat prompts available to the user. If the `llm_blueprint`, `playground`, or `chat`
        is specified then the results are restricted to the chat prompts associated with that
        entity.

        Parameters
        ----------
        llm_blueprint : Optional[Union[LLMBlueprint, str]], optional
            The returned chat prompts are filtered to those associated with a specific LLM blueprint
            if it is specified. Accepts either `LLMBlueprint` or LLM blueprint ID.
        playground : Optional[Union[Playground, str]], optional
            The returned chat prompts are filtered to those associated with a specific playground
            if it is specified. Accepts either `Playground` or playground ID.
        chat : Optional[Union[Chat, str]], optional
            The returned chat prompts are filtered to those associated with a specific chat
            if it is specified. Accepts either `Chat` or chat ID.

        Returns
        -------
        chat_prompts : list[ChatPrompt]
            A list of chat prompts available to the user.
        """
        params = {
            "llm_blueprint_id": get_entity_id(llm_blueprint) if llm_blueprint else None,
            "playground_id": get_entity_id(playground) if playground else None,
            "chat_id": get_entity_id(chat) if chat else None,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def delete(self) -> None:
        """
        Delete the single chat prompt.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)

    def create_llm_blueprint(self, name: str, description: str = "") -> LLMBlueprint:
        """
        Create a new LLM blueprint from an existing chat prompt.

        Parameters
        ----------
        name : str
            LLM blueprint name.
        description : str, optional
            Description of the LLM blueprint, by default "".

        Returns
        -------
        llm_blueprint : LLMBlueprint
            The created LLM blueprint.
        """
        payload = {
            "chat_prompt_id": self.id,
            "name": name,
            "description": description,
        }

        url = f"{self._client.domain}/{LLMBlueprint._path}/fromChatPrompt/"
        r_data = self._client.post(url, data=payload)
        return LLMBlueprint.from_server_data(r_data.json())



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot._compat import String
from datarobot.enums import DATA_STORE_TABLE_TYPE, DataStoreListTypes, DataStoreTypes
from datarobot.models.api_object import APIObject, ServerDataType
from datarobot.models.credential import CredentialDataSchema
from datarobot.models.sharing import SharingAccess, SharingRole
from datarobot.utils import deprecated, deprecation_warning, from_api, parse_time, to_api
from datarobot.utils.pagination import unpaginate

field_converter = t.Dict(
    {t.Key("id"): String(), t.Key("name"): String(), t.Key("value"): String()}
).ignore_extra("*")
_data_store_params_converter = t.Dict(
    {
        t.Key("driver_id", optional=True): t.Or(String(), t.Null()),
        t.Key("connector_id", optional=True): t.Or(String(), t.Null()),
        t.Key("jdbc_url", optional=True): t.Or(String(), t.Null()),
        t.Key("fields", optional=True): t.Or(t.List(field_converter), t.Null()),
    }
).ignore_extra("*")

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class TestResponse(TypedDict):
        message: str

    class SchemasResponse(TypedDict):
        schemas: List[str]
        catalogs: Optional[List[str]]
        catalog: str

    class TableDescription(TypedDict):
        catalog: Optional[str]
        name: str
        schema: Optional[str]
        type: DATA_STORE_TABLE_TYPE

    class TablesResponse(TypedDict):
        catalog: str
        tables: List[TableDescription]


class DataStoreParameters:
    """A data store's parameters'

    Attributes
    ----------
    driver_id : str
        Optional. The identifier of the data driver if the type is one of DataStoreTypes.DR_DATABASE_V1
        or DataStoreTypes.JDBC.
    jdbc_url : str
        Optional. The full JDBC URL (for example: `jdbc:postgresql://my.dbaddress.org:5432/my_db`).
    fields: list
        Optional. If the type is `dr-database-v1`, then the fields specify the configuration.
    connector_id: str
        Optional. The connector identifier if the type is DataStoreTypes.DR_CONNECTOR_V1.
    """

    def __init__(
        self,
        driver_id: Optional[str],
        jdbc_url: Optional[str],
        fields: Optional[List[Dict[str, str]]] = None,
        connector_id: Optional[str] = None,
    ):
        _data_store_params_converter.check(
            {
                "driver_id": driver_id,
                "jdbc_url": jdbc_url,
                "fields": fields,
                "connector_id": connector_id,
            }
        )
        self.driver_id = driver_id
        self.connector_id = connector_id
        self.jdbc_url = jdbc_url
        self.fields = fields

    def collect_payload(self) -> Dict[str, Any]:
        """Build a dict of the parameters to send to the server"""
        dat: Dict[str, Any] = {}
        if self.driver_id:
            dat["driver_id"] = self.driver_id
        if self.connector_id:
            dat["connector_id"] = self.connector_id
        if self.jdbc_url:
            dat["jdbc_url"] = self.jdbc_url
        if self.fields:
            dat["fields"] = self.fields
        return dat


class DataStore(APIObject):
    """A data store. Represents database

    Attributes
    ----------
    id : str
        The id of the data store.
    data_store_type : str
        The type of data store.
    canonical_name : str
        The user-friendly name of the data store.
    creator : str
        The id of the user who created the data store.
    updated : datetime.datetime
        The time of the last update
    params : DataStoreParameters
        A list specifying data store parameters.
    role : str
        Your access role for this data store.
    """

    _path = "externalDataStores/"
    _converter = t.Dict(
        {
            t.Key("id", optional=True) >> "data_store_id": String(),
            t.Key("type") >> "data_store_type": String(),
            t.Key("canonical_name"): String(),
            t.Key("creator"): String(),
            t.Key("params"): _data_store_params_converter,
            t.Key("updated"): parse_time,
            t.Key("role"): String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        data_store_id: Optional[str] = None,
        data_store_type: Optional[str] = None,
        canonical_name: Optional[str] = None,
        creator: Optional[str] = None,
        updated: Optional[datetime] = None,
        params: Optional[DataStoreParameters] = None,
        role: Optional[str] = None,
    ):
        self._id = data_store_id
        self._type = data_store_type
        self.canonical_name = canonical_name
        self._creator = creator
        self._updated = updated
        self.params = params
        self.role = role

    @classmethod
    def list(
        cls, typ: Optional[Union[str, DataStoreListTypes]] = None, name: Optional[str] = None
    ) -> List[DataStore]:
        """
        Returns list of available data stores.

        Parameters
        ----------
        typ : str
            If specified, filters by specified data store type. If not specified, the default
            is DataStoreListTypes.JDBC.
        name: str
            If specified, filters by data store names that match or contain this name.
            The search is case-insensitive.

        Returns
        -------
        data_stores : list of DataStore instances
            contains a list of available data stores.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_stores = dr.DataStore.list()
            >>> data_stores
            [DataStore('Demo'), DataStore('Airlines')]
        """
        params = {}
        if typ:
            params["type"] = typ
        if name:
            params["name"] = name

        if params:
            r_data = cls._client.get(cls._path, params=params).json()
        else:
            r_data = cls._client.get(cls._path).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls, data_store_id: str) -> DataStore:
        """
        Gets the data store.

        Parameters
        ----------
        data_store_id : str
            the identifier of the data store.

        Returns
        -------
        data_store : DataStore
            the required data store.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5a8ac90b07a57a0001be501e')
            >>> data_store
            DataStore('Demo')
        """
        return cls.from_location(f"{cls._path}{data_store_id}/")

    @classmethod
    def create(
        cls,
        data_store_type: Union[str, DataStoreTypes],
        canonical_name: str,
        driver_id: Optional[str] = None,
        jdbc_url: Optional[str] = None,
        fields: Optional[List[Dict[str, str]]] = None,
        connector_id: Optional[str] = None,
    ) -> DataStore:
        """
        Creates the data store.

        Parameters
        ----------
        data_store_type : str or DataStoreTypes
            the type of data store.
        canonical_name : str
            the user-friendly name of the data store.
        driver_id : str
            Optional. The identifier of the DataDriver if data_store_type is DataStoreListTypes.JDBC or
            DataStoreListTypes.DR_DATABASE_V1.
        jdbc_url : str
            Optional. The full JDBC URL (for example: `jdbc:postgresql://my.dbaddress.org:5432/my_db`).
        fields: list
            Optional. If the type is `dr-database-v1`, then the fields specify the configuration.
        connector_id: str
            Optional. The identifier of the Connector if data_store_type is DataStoreListTypes.DR_CONNECTOR_V1
        Returns
        -------
        data_store : DataStore
            the created data store.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.create(
            ...     data_store_type='jdbc',
            ...     canonical_name='Demo DB',
            ...     driver_id='5a6af02eb15372000117c040',
            ...     jdbc_url='jdbc:postgresql://my.db.address.org:5432/perftest'
            ... )
            >>> data_store
            DataStore('Demo DB')
        """
        payload = {
            "type": str(data_store_type),
            "canonicalName": canonical_name,
            "params": DataStoreParameters(
                driver_id=driver_id, jdbc_url=jdbc_url, fields=fields, connector_id=connector_id
            ).collect_payload(),
        }
        return cls.from_server_data(cls._client.post(cls._path, data=payload).json())

    def update(
        self,
        canonical_name: Optional[str] = None,
        driver_id: Optional[str] = None,
        connector_id: Optional[str] = None,
        jdbc_url: Optional[str] = None,
        fields: Optional[List[Dict[str, str]]] = None,
    ) -> None:
        """
        Updates the data store.

        Parameters
        ----------
        canonical_name : str
            optional, the user-friendly name of the data store.
        driver_id : str
            Optional. The identifier of the DataDriver. if the type is one of DataStoreTypes.DR_DATABASE_V1
            or DataStoreTypes.JDBC.
        connector_id : str
            Optional. The identifier of the Connector. if the type is DataStoreTypes.DR_CONNECTOR_V1.
        jdbc_url : str
            Optional. The full JDBC URL (for example: `jdbc:postgresql://my.dbaddress.org:5432/my_db`).
        fields: list
            Optional. If the type is `dr-database-v1`, then the fields specify the configuration.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store
            DataStore('Demo DB')
            >>> data_store.update(canonical_name='Demo DB updated')
            >>> data_store
            DataStore('Demo DB updated')
        """
        params = DataStoreParameters(
            driver_id=driver_id or self.params.driver_id,  # type: ignore[union-attr]
            connector_id=connector_id or self.params.connector_id,  # type: ignore[union-attr]
            jdbc_url=jdbc_url or self.params.jdbc_url,  # type: ignore[union-attr]
            fields=fields or self.params.fields,  # type: ignore[union-attr]
        ).collect_payload()
        # if we are updating fields, then we cannot include driver_id or connector_id
        if params.get("fields"):
            params.pop("driver_id", None)
            params.pop("connector_id", None)

        payload = {
            "canonicalName": canonical_name or self.canonical_name,
            "params": params,
        }
        r_data = self._client.patch(f"{self._path}{self.id}/", data=payload).json()
        self.canonical_name = r_data["canonicalName"]
        self.params = DataStoreParameters(
            r_data["params"].get("driverId"),
            r_data["params"].get("jdbcUrl"),
            r_data["params"].get("fields"),
            r_data["params"].get("connectorId"),
        )

    def delete(self) -> None:
        """Removes the DataStore"""
        self._client.delete(f"{self._path}{self.id}/")

    def test(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        credential_id: Optional[str] = None,
        use_kerberos: Optional[bool] = None,
        credential_data: Optional[Dict[str, str]] = None,
    ) -> TestResponse:
        """
        Tests database connection.

        .. versionchanged:: v3.2
           Added `credential_id`, `use_kerberos` and `credential_data` optional params and made
           `username` and `password` optional.

        Parameters
        ----------
        username : str
            optional, the username for database authentication.
        password : str
            optional, the password for database authentication. The password is encrypted
            at server side and never saved / stored
        credential_id : str
            optional, id of the set of credentials to use instead of username and password
        use_kerberos : bool
            optional, whether to use Kerberos for data store authentication
        credential_data : dict
            optional, the credentials to authenticate with the database, to use instead of
            user/password or credential ID

        Returns
        -------
        message : dict
            message with status.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store.test(username='db_username', password='db_password')
            {'message': 'Connection successful'}
        """
        payload = {
            "user": username,
            "password": password,
            "credential_id": credential_id,
            "use_kerberos": use_kerberos,
        }
        if credential_data:
            payload["credential_data"] = CredentialDataSchema(credential_data)
        return self._client.post(f"{self._path}{self.id}/test/", data=to_api(payload)).json()  # type: ignore[no-any-return] # noqa: E501

    def schemas(self, username: str, password: str) -> SchemasResponse:
        """
        Returns list of available schemas.

        Parameters
        ----------
        username : str
            the username for database authentication.
        password : str
            the password for database authentication. The password is encrypted
            at server side and never saved / stored

        Returns
        -------
        response : dict
            dict with database name and list of str - available schemas

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store.schemas(username='db_username', password='db_password')
            {'catalog': 'perftest', 'schemas': ['demo', 'information_schema', 'public']}
        """
        payload = {"user": username, "password": password}
        return self._client.post(f"{self._path}{self.id}/schemas/", data=payload).json()  # type: ignore[no-any-return]

    def tables(self, username: str, password: str, schema: Optional[str] = None) -> TablesResponse:
        """
        Returns list of available tables in schema.

        Parameters
        ----------
        username : str
            optional, the username for database authentication.
        password : str
            optional, the password for database authentication. The password is encrypted
            at server side and never saved / stored
        schema : str
            optional, the schema name.

        Returns
        -------
        response : dict
            dict with catalog name and tables info

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> data_store = dr.DataStore.get('5ad5d2afef5cd700014d3cae')
            >>> data_store.tables(username='db_username', password='db_password', schema='demo')
            {'tables': [{'type': 'TABLE', 'name': 'diagnosis', 'schema': 'demo'}, {'type': 'TABLE',
            'name': 'kickcars', 'schema': 'demo'}, {'type': 'TABLE', 'name': 'patient',
            'schema': 'demo'}, {'type': 'TABLE', 'name': 'transcript', 'schema': 'demo'}],
            'catalog': 'perftest'}
        """
        payload = {"schema": schema, "user": username, "password": password}
        return self._client.post(f"{self._path}{self.id}/tables/", data=payload).json()  # type: ignore[no-any-return]

    @classmethod
    def from_server_data(  # type: ignore[override]
        cls, data: ServerDataType, keep_attrs: Optional[List[str]] = None
    ) -> DataStore:
        converted_data = cls._converter.check(from_api(data))
        params = converted_data.pop("params")
        converted_data["params"] = DataStoreParameters(
            params.get("driver_id"),
            params.get("jdbc_url"),
            params.get("fields"),
            params.get("connector_id"),
        )
        return cls(**converted_data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.canonical_name or self.id}')"

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def creator(self) -> Optional[str]:
        return self._creator

    @property
    def type(self) -> Optional[str]:
        return self._type

    @property
    def updated(self) -> Optional[datetime]:
        return self._updated

    @deprecated(
        deprecated_since_version="v3.2",
        will_remove_version="v3.4",
        message="Please use get_shared_roles instead.",
    )
    def get_access_list(self) -> List[SharingAccess]:
        """Retrieve what users have access to this data store

        .. versionadded:: v2.14

        Returns
        -------
        list of :class:`SharingAccess <datarobot.SharingAccess>`
        """
        url = f"{self._path}{self.id}/accessControl/"
        return [
            SharingAccess.from_server_data(datum) for datum in unpaginate(url, {}, self._client)
        ]

    def get_shared_roles(self) -> List[SharingRole]:
        """Retrieve what users have access to this data store

        .. versionadded:: v3.2

        Returns
        -------
        list of :class:`SharingRole <datarobot.models.sharing.SharingRole>`
        """
        url = f"{self._path}{self.id}/sharedRoles/"
        return [SharingRole.from_server_data(datum) for datum in unpaginate(url, {}, self._client)]

    def share(self, access_list: Union[List[SharingAccess], List[SharingRole]]) -> None:
        """Modify the ability of users to access this data store

        .. versionadded:: v2.14

        Parameters
        ----------
        access_list : list of :class:`SharingRole <datarobot.models.sharing.SharingRole>`
            the modifications to make.

        Raises
        ------
        datarobot.ClientError :
            if you do not have permission to share this data store, if the user you're sharing with
            doesn't exist, if the same user appears multiple times in the access_list, or if these
            changes would leave the data store without an owner.

        Examples
        --------
        The :class:`SharingRole <datarobot.models.sharing.SharingRole>` class is needed in order to
        share a Data Store with one or more users.

        For example, suppose you had a list of user IDs you wanted to share this DataStore with. You could use
        a loop to generate a list of :class:`SharingRole <datarobot.models.sharing.SharingRole>` objects for them,
        and bulk share this Data Store.

        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.sharing import SharingRole
            >>> from datarobot.enums import SHARING_ROLE, SHARING_RECIPIENT_TYPE
            >>>
            >>> user_ids = ["60912e09fd1f04e832a575c1", "639ce542862e9b1b1bfa8f1b", "63e185e7cd3a5f8e190c6393"]
            >>> sharing_roles = []
            >>> for user_id in user_ids:
            ...     new_sharing_role = SharingRole(
            ...         role=SHARING_ROLE.CONSUMER,
            ...         share_recipient_type=SHARING_RECIPIENT_TYPE.USER,
            ...         id=user_id,
            ...         can_share=True,
            ...     )
            ...     sharing_roles.append(new_sharing_role)
            >>> dr.DataStore.get('my-data-store-id').share(access_list)

        Similarly, a :class:`SharingRole <datarobot.models.sharing.SharingRole>` instance can be used to
        remove a user's access if the ``role`` is set to ``SHARING_ROLE.NO_ROLE``, like in this example:

        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.sharing import SharingRole
            >>> from datarobot.enums import SHARING_ROLE, SHARING_RECIPIENT_TYPE
            >>>
            >>> user_to_remove = "foo.bar@datarobot.com"
            ... remove_sharing_role = SharingRole(
            ...     role=SHARING_ROLE.NO_ROLE,
            ...     share_recipient_type=SHARING_RECIPIENT_TYPE.USER,
            ...     username=user_to_remove,
            ...     can_share=False,
            ... )
            >>> dr.DataStore.get('my-data-store-id').share(roles=[remove_sharing_role])
        """
        if any(isinstance(access, SharingAccess) for access in access_list):
            if not (all(isinstance(access, SharingAccess) for access in access_list)):
                raise ValueError("Please use either all SharingRole or all SharingAccess objects.")
            deprecation_warning(
                subject="SharingAccess",
                deprecated_since_version="v3.2",
                will_remove_version="v3.4",
                message="Please use SharingRole objects instead of SharingAccess objects.",
            )
            payload = {"data": [access.collect_payload() for access in access_list]}
            self._client.patch(
                f"{self._path}{self.id}/accessControl/", data=payload, keep_attrs={"role"}
            )
        else:
            formatted_roles = [access.collect_payload() for access in access_list]
            payload = {"roles": formatted_roles, "operation": "updateRoles"}  # type: ignore[dict-item]
            self._client.patch(f"{self._path}{self.id}/sharedRoles/", data=payload)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Optional

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject


class PrimeFile(APIObject):
    """Represents an executable file available for download of the code for a DataRobot Prime model

    Attributes
    ----------
    id : str
        the id of the PrimeFile
    project_id : str
        the id of the project this PrimeFile belongs to
    parent_model_id : str
        the model being approximated by this PrimeFile
    model_id : str
        the prime model this file represents
    ruleset_id : int
        the ruleset being used in this PrimeFile
    language : str
        the language of the code in this file - see enums.LANGUAGE for possibilities
    is_valid : bool
        whether the code passed basic validation
    """

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("project_id"): String(),
            t.Key("parent_model_id"): String(),
            t.Key("model_id"): String(),
            t.Key("ruleset_id"): Int(),
            t.Key("language"): String(),
            t.Key("is_valid"): t.Bool(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        project_id: Optional[str] = None,
        parent_model_id: Optional[str] = None,
        model_id: Optional[str] = None,
        ruleset_id: Optional[int] = None,
        language: Optional[str] = None,
        is_valid: Optional[bool] = None,
    ) -> None:
        self.id = id
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id
        self.ruleset_id = ruleset_id
        self.language = language
        self.is_valid = is_valid

    @classmethod
    def get(cls, project_id: str, file_id: str) -> PrimeFile:
        url = f"projects/{project_id}/primeFiles/{file_id}/"
        return cls.from_location(url)

    def download(self, filepath: str) -> None:
        """Download the code and save it to a file

        Parameters
        ----------
        filepath: string
            the location to save the file to
        """
        url = f"projects/{self.project_id}/primeFiles/{self.id}/download/"
        response = self._client.get(url)
        with open(filepath, mode="wb") as out_file:
            out_file.write(response.content)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import dateutil
import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT, ExportStatus
from datarobot.models.api_object import APIObject
from datarobot.models.dataset import Dataset
from datarobot.utils import datetime_to_string, from_api
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class Period(TypedDict, total=False):
        start: datetime
        end: datetime

    class ExportError(TypedDict, total=False):
        message: str

    class ExportDataset(TypedDict, total=False):
        id: str
        name: str

    class ExportBatches(TypedDict, total=False):
        batch_id: str
        batch_name: str


class PredictionDataExport(APIObject):
    """A prediction data export.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the prediction data export.
    model_id: str
        The ID of the model (or null if not specified).
    created_at: datetime
        Prediction data export creation timestamp.
    period: Period
        A prediction data time range definition.
    status: ExportStatus
       A prediction data export processing state.
    error: ExportError
       Error description, appears when prediction data export job failed (status is FAILED).
    batches: ExportBatches
       Metadata associated with exported batch.
    deployment_id: str
        The ID of the deployment.
    """

    _path = "deployments/{}/predictionDataExports/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("model_id"): t.String(),
            t.Key("created_at"): t.String() >> dateutil.parser.parse,
            t.Key("period"): t.Dict(
                {
                    t.Key("start"): t.String >> dateutil.parser.parse,
                    t.Key("end"): t.String >> dateutil.parser.parse,
                }
            ),
            t.Key("status"): t.String(),
            t.Key("error", optional=True, default=None): t.Or(
                t.Dict().allow_extra("*"),
                t.Null,
            ),
            t.Key("data", optional=True): t.Or(
                t.List(t.Dict().allow_extra("*")),
                t.Null,
            ),
            t.Key("batches", optional=True, default=None): t.Or(
                t.Dict().allow_extra("*"),
                t.Null,
            ),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        period: Period,
        created_at: datetime,
        model_id: str,
        status: ExportStatus,
        data: Optional[List[ExportDataset]] = None,
        error: Optional[ExportError] = None,
        batches: Optional[ExportBatches] = None,
        deployment_id: Optional[str] = None,
    ):
        self.id = id
        self.period = period
        self.created_at = created_at
        self.model_id = model_id
        self.status = status
        self.data = data
        self.error = error
        self.batches = batches
        self.deployment_id = deployment_id

    def __repr__(self) -> str:
        return "PredictionDataExport({})".format(self.id)

    @classmethod
    def list(
        cls,
        deployment_id: str,
        status: Optional[ExportStatus] = None,
        model_id: Optional[str] = None,
        batch: Optional[bool] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 100,
    ) -> List[PredictionDataExport]:
        """Retrieve a list of prediction data exports.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        model_id: Optional[str]
            The ID of the model used for prediction data export.
        status: Optional[ExportStatus]
            A prediction data export processing state.
        batch: Optional[bool]
            If true, only return batch exports.
            If false, only return real-time exports.
            If not provided, return both real-time and batch exports.
        limit: Optional[int]
            The maximum number of objects to return. The default is 100 (0 means no limit).
        offset: Optional[int]
            The starting offset of the results. The default is 0.

        Returns
        -------
        prediction_data_exports: list
            A list of PredictionDataExport objects.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import PredictionDataExport

            prediction_data_exports = PredictionDataExport.list(deployment_id='5c939e08962d741e34f609f0')
        """
        path = cls._path.format(deployment_id)

        params: Dict[str, Any] = {}
        if model_id:
            params["modelId"] = model_id
        if status:
            params["status"] = status
        if batch:
            params["batch"] = batch

        if limit == 0:  # unlimited results
            r_data = unpaginate(path, params, cls._client)
        else:
            params["offset"] = offset
            params["limit"] = limit
            response = cls._client.get(path, params=params).json()
            r_data = response["data"]

        prediction_data_exports = [
            cls.from_server_data(from_api(d, keep_null_keys=True)) for d in r_data
        ]
        for data_export in prediction_data_exports:
            data_export.deployment_id = deployment_id
        return prediction_data_exports

    @classmethod
    def get(
        cls,
        deployment_id: str,
        export_id: str,
    ) -> PredictionDataExport:
        """Retrieve a single prediction data export.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        export_id: str
            The ID of the prediction data export.

        Returns
        -------
        prediction_data_export: PredictionDataExport
            A prediction data export.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import PredictionDataExport

            prediction_data_export = PredictionDataExport.get(
                deployment_id='5c939e08962d741e34f609f0', export_id='65fbe59aaa3f847bd5acc75b'
                )
        """
        path = "{}{}/".format(cls._path.format(deployment_id), export_id)
        data = cls._client.get(path).json()
        case_converted = from_api(data, keep_null_keys=True)
        prediction_data_export = cls.from_data(case_converted)
        prediction_data_export.deployment_id = deployment_id
        return prediction_data_export

    @classmethod
    def create(
        cls,
        deployment_id: str,
        start: Union[datetime, str],
        end: Union[datetime, str],
        model_id: Optional[str] = None,
        batch_ids: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> PredictionDataExport:
        """Create a deployment prediction data export.
        Waits until ready and fetches PredictionDataExport after the export finishes. This method is blocking.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        start: Union[datetime, str]
            Inclusive start of the time range.
        end: Union[datetime, str]
            Exclusive end of the time range.
        model_id: Optional[str]
            The ID of the model.
        batch_ids: Optional[List[str]]
            IDs of batches to export. Null for real-time data exports.
        max_wait: int,
            Seconds to wait for successful resolution.

        Returns
        -------
        prediction_data_export: PredictionDataExport
            A prediction data export.

        Examples
        --------
        .. code-block:: python

            from datetime import datetime, timedelta
            from datarobot.models.deployment import PredictionDataExport

            now=datetime.now()
            prediction_data_export = PredictionDataExport.create(
                deployment_id='5c939e08962d741e34f609f0', start=now - timedelta(days=7), end=now
                )
        """
        payload = {
            "start": start if isinstance(start, str) else datetime_to_string(start),
            "end": end if isinstance(end, str) else datetime_to_string(end),
            "batchIds": batch_ids,
        }
        if model_id:
            payload["modelId"] = model_id
        if batch_ids:
            payload["batchIds"] = batch_ids
        path = cls._path.format(deployment_id)
        response = cls._client.post(path, json=payload)
        retrieve_url = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        response = cls._client.get(retrieve_url)
        export_id = response.json()["id"]
        return cls.get(deployment_id, export_id)

    def fetch_data(self) -> List[Dataset]:
        """Return data from prediction export as datarobot Dataset.

        Returns
        -------
        prediction_datasets: List[Dataset]
            List of datasets for a given export, most often it is just one.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import PredictionDataExport

            prediction_data_export = PredictionDataExport.get(
                deployment_id='5c939e08962d741e34f609f0', export_id='65fbe59aaa3f847bd5acc75b'
                )
            prediction_datasets = prediction_data_export.fetch_data()
        """
        if not self.data:
            raise ValueError("No datasets found for prediction export.")
        datasets = [Dataset.get(item["id"]) for item in self.data]
        return datasets


class ActualsDataExport(APIObject):
    """An actuals data export.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the actuals data export.
    model_id: str
        The ID of the model (or null if not specified).
    created_at: datetime
        Actuals data export creation timestamp.
    period: Period
        A actuals data time range definition.
    status: ExportStatus
       A data export processing state.
    error: ExportError
       Error description, appears when actuals data export job failed (status is FAILED).
    only_matched_predictions: bool
        If true, exports actuals with matching predictions only.
    deployment_id: str
        The ID of the deployment.
    """

    _path = "deployments/{}/actualsDataExports/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("model_id"): t.String(),
            t.Key("created_at"): t.String() >> dateutil.parser.parse,
            t.Key("period"): t.Dict(
                {
                    t.Key("start"): t.String >> dateutil.parser.parse,
                    t.Key("end"): t.String >> dateutil.parser.parse,
                }
            ),
            t.Key("status"): t.String(),
            t.Key("error", optional=True, default=None): t.Or(
                t.Dict().allow_extra("*"),
                t.Null,
            ),
            t.Key("data", optional=True): t.Or(
                t.List(t.Dict().allow_extra("*")),
                t.Null,
            ),
            t.Key("only_matched_predictions", optional=True, default=True): t.Bool(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        period: Period,
        created_at: datetime,
        model_id: str,
        status: ExportStatus,
        data: Optional[List[ExportDataset]] = None,
        error: Optional[ExportError] = None,
        only_matched_predictions: Optional[bool] = None,
        deployment_id: Optional[str] = None,
    ):
        self.id = id
        self.period = period
        self.created_at = created_at
        self.model_id = model_id
        self.status = status
        self.data = data
        self.error = error
        self.only_matched_predictions = only_matched_predictions
        self.deployment_id = deployment_id

    def __repr__(self) -> str:
        return "ActualsDataExport({})".format(self.id)

    @classmethod
    def list(
        cls,
        deployment_id: str,
        status: Optional[ExportStatus] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 100,
    ) -> List[ActualsDataExport]:
        """Retrieve a list of actuals data exports.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        status: Optional[ExportStatus]
            Actuals data export processing state.
        limit: Optional[int]
            The maximum number of objects to return. The default is 100 (0 means no limit).
        offset: Optional[int]
            The starting offset of the results. The default is 0.

        Returns
        -------
        actuals_data_exports: list
            A list of ActualsDataExport objects.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import ActualsDataExport

            actuals_data_exports = ActualsDataExport.list(deployment_id='5c939e08962d741e34f609f0')
        """
        path = cls._path.format(deployment_id)

        params: Dict[str, Any] = {}
        if status:
            params["status"] = status

        if limit == 0:  # unlimited results
            r_data = unpaginate(path, params, cls._client)
        else:
            params["offset"] = offset
            params["limit"] = limit
            response = cls._client.get(path, params=params).json()
            r_data = response["data"]

        actuals_data_exports = [
            cls.from_server_data(from_api(d, keep_null_keys=True)) for d in r_data
        ]
        for data_export in actuals_data_exports:
            data_export.deployment_id = deployment_id
        return actuals_data_exports

    @classmethod
    def get(
        cls,
        deployment_id: str,
        export_id: str,
    ) -> ActualsDataExport:
        """Retrieve a single actuals data export.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        export_id: str
            The ID of the actuals data export.

        Returns
        -------
        actuals_data_export: ActualsDataExport
            An actuals data export.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import ActualsDataExport

            actuals_data_export = ActualsDataExport.get(
                deployment_id='5c939e08962d741e34f609f0', export_id='65fb0a6c9bb187781cfdea36'
                )
        """
        path = "{}{}/".format(cls._path.format(deployment_id), export_id)
        data = cls._client.get(path).json()
        case_converted = from_api(data, keep_null_keys=True)
        actuals_data_export = cls.from_data(case_converted)
        actuals_data_export.deployment_id = deployment_id
        return actuals_data_export

    @classmethod
    def create(
        cls,
        deployment_id: str,
        start: Union[datetime, str],
        end: Union[datetime, str],
        model_id: Optional[str] = None,
        only_matched_predictions: Optional[bool] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> ActualsDataExport:
        """Create a deployment actuals data export.
        Waits until ready and fetches ActualsDataExport after the export finishes. This method is blocking.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        start: Union[datetime, str]
            Inclusive start of the time range.
        end: Union[datetime, str]
            Exclusive end of the time range.
        model_id: Optional[str]
            The ID of the model.
        only_matched_predictions: Optional[bool]
            If true, exports actuals with matching predictions only.
        max_wait: int
            Seconds to wait for successful resolution.

        Returns
        -------
        actuals_data_export: ActualsDataExport
            An actuals data export.

        Examples
        --------
        .. code-block:: python

            from datetime import datetime, timedelta
            from datarobot.models.deployment import ActualsDataExport

            now=datetime.now()
            actuals_data_export = ActualsDataExport.create(
                deployment_id='5c939e08962d741e34f609f0', start=now - timedelta(days=7), end=now
                )
        """
        payload: Dict[str, Any] = {
            "start": start if isinstance(start, str) else datetime_to_string(start),
            "end": end if isinstance(end, str) else datetime_to_string(end),
        }
        if model_id:
            payload["modelId"] = model_id
        if only_matched_predictions:
            payload["onlyMatchedPredictions"] = only_matched_predictions
        path = cls._path.format(deployment_id)
        response = cls._client.post(path, json=payload)
        retrieve_url = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        response = cls._client.get(retrieve_url)
        export_id = response.json()["id"]
        return cls.get(deployment_id, export_id)

    def fetch_data(self) -> List[Dataset]:
        """Return data from actuals export as datarobot Dataset.

        Returns
        -------
        actuals_datasets: List[Dataset]
            List of datasets for a given export, most often it is just one.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import ActualsDataExport

            actuals_data_export = ActualsDataExport.get(
                deployment_id='5c939e08962d741e34f609f0', export_id='65fb0a6c9bb187781cfdea36'
                )
            actuals_datasets = actuals_data_export.fetch_data()
        """
        if not self.data:
            raise ValueError("No datasets found for actuals export.")
        datasets = [Dataset.get(item["id"]) for item in self.data]
        return datasets


class TrainingDataExport(APIObject):
    """A training data export.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the training data export.
    model_id: str
        The ID of the model (or null if not specified).
    model_package_id: str
        The ID of the model package.
    created_at: datetime
        Training data export creation timestamp.
    deployment_id: str
        The ID of the deployment.
    """

    _path = "deployments/{}/trainingDataExports/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("model_id"): t.String(),
            t.Key("model_package_id"): t.String(),
            t.Key("created_at"): t.String() >> dateutil.parser.parse,
            t.Key("data", optional=True): t.Or(
                t.Dict().allow_extra("*"),
                t.Null,
            ),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        created_at: datetime,
        model_id: str,
        model_package_id: str,
        data: Optional[ExportDataset] = None,
        deployment_id: Optional[str] = None,
    ):
        self.id = id
        self.data = data
        self.created_at = created_at
        self.model_id = model_id
        self.model_package_id = model_package_id
        self.deployment_id = deployment_id

    def __repr__(self) -> str:
        return "TrainingDataExport({})".format(self.id)

    @classmethod
    def list(
        cls,
        deployment_id: str,
    ) -> List[TrainingDataExport]:
        """Retrieve a list of successful training data exports.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.

        Returns
        -------
        training_data_exports: list
            A list of TrainingDataExport objects.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import TrainingDataExport

            training_data_exports = TrainingDataExport.list(deployment_id='5c939e08962d741e34f609f0')
        """
        path = cls._path.format(deployment_id)
        data = cls._client.get(path).json()["data"]
        training_data_exports = [
            cls.from_server_data(from_api(d, keep_null_keys=True)) for d in data
        ]
        for data_export in training_data_exports:
            data_export.deployment_id = deployment_id
        return training_data_exports

    @classmethod
    def get(
        cls,
        deployment_id: str,
        export_id: str,
    ) -> TrainingDataExport:
        """Retrieve a single training data export.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        export_id: str
            The ID of the training data export.

        Returns
        -------
        training_data_export: TrainingDataExport
            A training data export.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import TrainingDataExport
            training_data_export = TrainingDataExport.get(
                deployment_id='5c939e08962d741e34f609f0', export_id='65fbf2356124f1daa3acc522'
                )
        """
        path = "{}{}/".format(cls._path.format(deployment_id), export_id)
        data = cls._client.get(path).json()
        case_converted = from_api(data, keep_null_keys=True)
        training_data_export = cls.from_data(case_converted)
        training_data_export.deployment_id = deployment_id
        return training_data_export

    @classmethod
    def create(
        cls,
        deployment_id: str,
        model_id: Optional[str] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> str:
        """Create a single training data export.
        Waits until ready and fetches TrainingDataExport after the export finishes. This method is blocking.

        Parameters
        ----------
        deployment_id: str
            The ID of the deployment.
        model_id: Optional[str]
            The ID of the model.
        max_wait: int
            Seconds to wait for successful resolution.

        Returns
        -------
        dataset_id: str
            A created dataset with training data.

         Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import TrainingDataExport
            dataset_id = TrainingDataExport.create(deployment_id='5c939e08962d741e34f609f0')
        """
        payload = {}
        if model_id:
            payload["modelId"] = model_id
        path = cls._path.format(deployment_id)
        response = cls._client.post(path, json=payload)
        retrieve_url = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait
        )
        response = cls._client.get(retrieve_url)
        dataset_id = response.json()["datasetId"]
        return dataset_id  # type: ignore[no-any-return]

    def fetch_data(self) -> Dataset:
        """Return data from training data export as datarobot Dataset.

        Returns
        -------
        training_dataset: Dataset
            A datasets for a given export.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.deployment import TrainingDataExport

            training_data_export = TrainingDataExport.get(
                deployment_id='5c939e08962d741e34f609f0', export_id='65fbf2356124f1daa3acc522'
                )
            training_data_export = training_data_export.fetch_data()
        """
        if not self.data:
            raise ValueError("No datasets found for training data export.")
        dataset = Dataset.get(self.data["id"])
        return dataset



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import cast, List, Optional

from requests import Response
import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate


class PayoffMatrix(APIObject):
    """Represents a Payoff Matrix, a costs/benefit scenario used for creating a profit curve.

    Attributes
    ----------
    project_id : str
        id of the project with which the payoff matrix is associated.
    id : str
        id of the payoff matrix.
    name : str
        User-supplied label for the payoff matrix.
    true_positive_value : float
        Cost or benefit of a true positive classification
    true_negative_value : float
        Cost or benefit of a true negative classification
    false_positive_value : float
        Cost or benefit of a false positive classification
    false_negative_value : float
        Cost or benefit of a false negative classification

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # create a payoff matrix
        payoff_matrix = dr.PayoffMatrix.create(
            project_id,
            name,
            true_positive_value=100,
            true_negative_value=10,
            false_positive_value=0,
            false_negative_value=-10,
        )

        # list available payoff matrices
        payoff_matrices = dr.PayoffMatrix.list(project_id)
        payoff_matrix = payoff_matrices[0]
    """

    _base_url = "projects/{}/payoffMatrices/"
    _payoff_matrix_url = _base_url + "{}/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("project_id"): String(),
            t.Key("name"): String(),
            t.Key("true_positive_value"): t.Float(),
            t.Key("true_negative_value"): t.Float(),
            t.Key("false_positive_value"): t.Float(),
            t.Key("false_negative_value"): t.Float(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id: str,
        id: str,
        name: Optional[str] = None,
        true_positive_value: Optional[float] = None,
        true_negative_value: Optional[float] = None,
        false_positive_value: Optional[float] = None,
        false_negative_value: Optional[float] = None,
    ) -> None:
        self.project_id = project_id
        self.id = id
        self.name = name
        self.true_positive_value = true_positive_value
        self.true_negative_value = true_negative_value
        self.false_positive_value = false_positive_value
        self.false_negative_value = false_negative_value

    def __repr__(self) -> str:
        template = (
            "{}(id={!r}, project_id={!r}, name={!r}, true_positive_value={!r}, "
            "true_negative_value={!r}, false_positive_value={!r}, "
            "false_negative_value={!r})"
        )
        return template.format(
            type(self).__name__,
            self.id,
            self.project_id,
            self.name,
            self.true_positive_value,
            self.true_negative_value,
            self.false_positive_value,
            self.false_negative_value,
        )

    @classmethod
    def create(
        cls,
        project_id: str,
        name: str,
        true_positive_value: Optional[float] = 1,
        true_negative_value: Optional[float] = 1,
        false_positive_value: Optional[float] = -1,
        false_negative_value: Optional[float] = -1,
    ) -> PayoffMatrix:
        """
        Create a payoff matrix associated with a specific project.

        Parameters
        ----------
        project_id : str
            id of the project with which the payoff matrix will be associated

        Returns
        -------
        payoff_matrix : :py:class:`PayoffMatrix <datarobot.models.PayoffMatrix>`
            The newly created payoff matrix
        """
        data = {
            "name": name,
            "true_positive_value": true_positive_value,
            "true_negative_value": true_negative_value,
            "false_positive_value": false_positive_value,
            "false_negative_value": false_negative_value,
        }

        url = cls._base_url.format(project_id)
        response = cls._client.post(url, data=data)
        return cls.from_server_data(response.json())

    @classmethod
    def list(cls, project_id: str) -> List[PayoffMatrix]:
        """
        Fetch all the payoff matrices for a project.

        Parameters
        ----------
        project_id : str
            id of the project
        Returns
        -------
        List of PayoffMatrix
            A list of :py:class:`PayoffMatrix <datarobot.models.PayoffMatrix>` objects
        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            initial_url=cls._base_url.format(project_id),
            initial_params=None,
            client=cls._client,
        )
        result = [cls.from_server_data(item) for item in data]
        return result

    @classmethod
    def get(cls, project_id: str, id: str) -> PayoffMatrix:
        """
        Retrieve a specified payoff matrix.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        id : str
            id of the payoff matrix

        Returns
        -------
        :py:class:`PayoffMatrix <datarobot.models.PayoffMatrix>` object representing specified
        payoff matrix

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return cls.from_location(f"/projects/{project_id}/payoffMatrices/{id}")

    @classmethod
    def update(
        cls,
        project_id: str,
        id: str,
        name: str,
        true_positive_value: float,
        true_negative_value: float,
        false_positive_value: float,
        false_negative_value: float,
    ) -> PayoffMatrix:
        """Update (replace) a payoff matrix. Note that all data fields are required.

        Parameters
        ----------
        project_id : str
            id of the project to which the payoff matrix belongs
        id : str
            id of the payoff matrix
        name : str
            User-supplied label for the payoff matrix
        true_positive_value : float
            True positive payoff value to use for the profit curve
        true_negative_value : float
            True negative payoff value to use for the profit curve
        false_positive_value : float
            False positive payoff value to use for the profit curve
        false_negative_value : float
            False negative payoff value to use for the profit curve

        Returns
        -------
        payoff_matrix
            PayoffMatrix with updated values

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = {
            "name": name,
            "truePositiveValue": true_positive_value,
            "trueNegativeValue": true_negative_value,
            "falsePositiveValue": false_positive_value,
            "falseNegativeValue": false_negative_value,
        }

        url = f"projects/{project_id}/payoffMatrices/{id}/"
        response = cls._client.put(url, json=data)
        return cls.from_server_data(response.json())

    @classmethod
    def delete(cls, project_id: str, id: str) -> Response:
        """
        Delete a specified payoff matrix.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        id : str
            id of the payoff matrix

        Returns
        -------
        response : requests.Response
            Empty response (204)

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._payoff_matrix_url.format(project_id, id)
        return cast(Response, cls._client.delete(url))



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""

  AbstractJob
  ^        ^
  |        |
  +        +
Job       AbstractSpecificJob
            ^             ^
            |             |
            +             +
        ModelJob   AbstractBatchJob
                    ^           ^
                    |           |
                    +           +
        BatchPredictionJob    BatchMonitoringJob


"""
from __future__ import annotations

import csv
import io
import os
import threading
import time
import typing
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from mypy_extensions import TypedDict
import pandas as pd
import requests
import trafaret as t

from datarobot._compat import Int, String
from datarobot.utils import recognize_sourcedata as orig_recognize_sourcedata

from .. import errors
from ..enums import AVAILABLE_STATEMENT_TYPES, IntakeAdapters, OutputAdapters, TrainingDataSubsets
from .dataset import Dataset
from .job import AbstractSpecificJob


class IntakeSettings(TypedDict, total=False):
    """Intake settings typed dict"""

    type: IntakeAdapters
    file: Optional[Union[str, pd.DataFrame, io.IOBase]]
    url: Optional[str]
    credential_id: Optional[str]
    data_store_id: Optional[str]
    query: Optional[str]
    table: Optional[str]
    schema: Optional[str]
    catalog: Optional[str]
    fetch_size: Optional[int]
    format: Optional[str]
    endpoint_url: Optional[str]
    dataset: Optional[Dataset]
    dataset_id: Optional[str]
    dataset_version_id: Optional[str]


class OutputSettings(TypedDict, total=False):
    """Output settings typed dict"""

    type: OutputAdapters
    path: Optional[str]
    url: Optional[str]
    credential_id: Optional[str]
    data_store_id: Optional[str]
    table: Optional[str]
    schema: Optional[str]
    catalog: Optional[str]
    statement_type: Optional[str]
    update_columns: Optional[List[str]]
    where_columns: Optional[List[str]]
    create_table_if_not_exists: Optional[bool]


class CsvSettings(TypedDict):
    delimiter: Optional[str]
    quotechar: Optional[str]
    encoding: Optional[str]


class Schedule(TypedDict):
    day_of_week: List[Union[int, str]]
    month: List[Union[int, str]]
    hour: List[Union[int, str]]
    minute: List[Union[int, str]]
    day_of_month: List[Union[int, str]]


def recognize_sourcedata(sourcedata: Any, default_fname: str) -> Dict[str, Any]:
    """Override the default recognize_sourcedata with one that converts
    DataFrame to io.BytesIO"""
    if not isinstance(sourcedata, pd.DataFrame):
        return orig_recognize_sourcedata(sourcedata, default_fname)
    # The original recognize_sourcedata will encode dataframes into StringIO
    # To pass this to requests.put, we need BytesIO - otherwise, requests will
    # try to encode it using latin-1 encoding (the default HTTP encoding),
    # which will crash when a charater not contained in latin-1 is used
    #
    # https://github.com/python/cpython/blob/b9797417315cc2d1700cb2d427685016d3380711/Lib/http/client.py#L1046
    buf = io.BytesIO()
    sourcedata.to_csv(buf, encoding="utf-8", index=False, quoting=csv.QUOTE_ALL)
    buf.seek(0)
    return {"filelike": buf, "fname": default_fname}


class AbstractBatchJob(AbstractSpecificJob):
    """
    Generic representation of batch job, that covers common functions/checks
    for predictions, and monitoring jobs.

    """

    _s3_settings = t.Dict(
        {
            t.Key("url"): String(),
            t.Key("credential_id", optional=True): String(),
            t.Key("endpoint_url", optional=True): String(),
            t.Key("format", optional=True): String(),
        }
    )

    _gcp_settings = t.Dict(
        {
            t.Key("url"): String(),
            t.Key("credential_id", optional=True): String(),
            t.Key("format", optional=True): String(),
        }
    )

    _azure_settings = t.Dict(
        {
            t.Key("url"): String(),
            t.Key("credential_id", optional=True): String(),
            t.Key("format", optional=True): String(),
        }
    )

    _dataset_intake_settings = t.Dict(
        {t.Key("dataset"): t.Type(Dataset), t.Key("dataset_version_id", optional=True): String()}
    )

    _dss_intake_settings = t.Dict(
        {
            t.Key("project_id"): String(),
            t.Key("dataset_id", optional=True): String(),
            t.Key("partition", optional=True): t.Enum(
                TrainingDataSubsets.HOLDOUT,
                TrainingDataSubsets.VALIDATION,
                TrainingDataSubsets.ALL_BACKTESTS,
            ),
        }
    )

    _jdbc_intake_settings = t.Dict(
        {
            t.Key("data_store_id"): String(),
            t.Key("query", optional=True): String(),
            t.Key("table", optional=True): String(),
            t.Key("schema", optional=True): String(),
            t.Key("catalog", optional=True): String(),
            t.Key("fetch_size", optional=True): Int(),
            t.Key("credential_id", optional=True): String(),
        }
    )

    _jdbc_output_settings = t.Dict(
        {
            t.Key("data_store_id"): String(),
            t.Key("table"): String(),
            t.Key("schema", optional=True): String(),
            t.Key("catalog", optional=True): String(),
            t.Key("statement_type"): t.Enum(
                AVAILABLE_STATEMENT_TYPES.INSERT,
                AVAILABLE_STATEMENT_TYPES.UPDATE,
                AVAILABLE_STATEMENT_TYPES.INSERT_UPDATE,
                AVAILABLE_STATEMENT_TYPES.CREATE_TABLE,
            ),
            t.Key("update_columns", optional=True): t.List(String),
            t.Key("where_columns", optional=True): t.List(String),
            t.Key("credential_id", optional=True): String(),
            t.Key("create_table_if_not_exists", optional=True): t.Bool(),
        }
    )

    _csv_settings = t.Dict(
        {
            t.Key("delimiter", optional=True): t.Atom("tab") | String(min_length=1, max_length=1),
            t.Key("quotechar", optional=True): String(),
            t.Key("encoding", optional=True): String(),
        }
    )

    @classmethod
    def validate_intake_settings(
        cls, input_intake_settings: Optional[IntakeSettings] = None
    ) -> IntakeSettings:
        """Validates intake settings based on type run specific trafaret check
        Creates a copy in order to avoid mutating input data

        :param input_intake_settings:
        :return: A validated copy of intake settings
        """
        if input_intake_settings is None:
            intake_settings = IntakeSettings({"type": IntakeAdapters.LOCAL_FILE})
        else:
            # avoid mutating the input argument
            intake_settings = input_intake_settings.copy()

        # Validate the intake settings
        if intake_settings.get("type") not in IntakeAdapters:
            raise ValueError(
                "Unsupported type parameter for intake_settings: {}".format(
                    intake_settings.get("type")
                )
            )

        elif intake_settings["type"] == IntakeAdapters.LOCAL_FILE:
            # This intake option requires us to upload the source
            # data ourselves

            if intake_settings.get("file") is None:
                raise ValueError(
                    "Missing source data. Either supply the `file` "
                    "parameter or switch to an intake option that does not "
                    "require it."
                )

        elif intake_settings["type"] == IntakeAdapters.S3:
            del intake_settings["type"]
            intake_settings = cls._s3_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.S3

        elif intake_settings["type"] == IntakeAdapters.GCP:
            del intake_settings["type"]
            intake_settings = cls._gcp_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.GCP

        elif intake_settings["type"] == IntakeAdapters.AZURE:
            del intake_settings["type"]
            intake_settings = cls._azure_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.AZURE

        elif intake_settings["type"] == IntakeAdapters.JDBC:
            del intake_settings["type"]
            intake_settings = cls._jdbc_intake_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.JDBC

        elif intake_settings["type"] == IntakeAdapters.DATASET:
            del intake_settings["type"]
            intake_settings = cls._dataset_intake_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.DATASET

            dataset = intake_settings["dataset"]
            intake_settings["dataset_id"] = dataset.id  # type: ignore[union-attr]
            if "dataset_version_id" not in intake_settings:
                intake_settings["dataset_version_id"] = dataset.version_id  # type: ignore[union-attr]

            del intake_settings["dataset"]

        elif intake_settings["type"] == IntakeAdapters.DSS:
            del intake_settings["type"]
            intake_settings = cls._dss_intake_settings.check(intake_settings)
            intake_settings["type"] = IntakeAdapters.DSS

        return intake_settings

    @classmethod
    def validate_output_settings(
        cls, input_output_settings: Optional[OutputSettings] = None
    ) -> OutputSettings:
        """Validates output settings based on type run specific trafaret check
        Creates a copy in order to avoid mutating input data

        :param input_output_settings:
        :return: A validated copy of output settings
        """
        if input_output_settings is None:
            output_settings = OutputSettings({"type": OutputAdapters.LOCAL_FILE})
        else:
            output_settings = input_output_settings.copy()

        # Validate the output settings

        if output_settings.get("type") not in OutputAdapters:
            raise ValueError(
                "Unsupported type parameter for output_settings: {}".format(
                    output_settings.get("type")
                )
            )
        elif output_settings["type"] == OutputAdapters.LOCAL_FILE:
            output_settings["path"] = output_settings.get("path")

        elif output_settings["type"] == OutputAdapters.S3:
            del output_settings["type"]
            output_settings = cls._s3_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.S3

        elif output_settings["type"] == OutputAdapters.GCP:
            del output_settings["type"]
            output_settings = cls._gcp_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.GCP

        elif output_settings["type"] == OutputAdapters.AZURE:
            del output_settings["type"]
            output_settings = cls._azure_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.AZURE

        elif output_settings["type"] == OutputAdapters.JDBC:
            del output_settings["type"]
            output_settings = cls._jdbc_output_settings.check(output_settings)
            output_settings["type"] = OutputAdapters.JDBC

        return output_settings

    @classmethod
    def _upload_intake_file(
        cls,
        intake_file: Dict[str, Any],
        intake_settings: IntakeSettings,
        job_response: Dict[str, Any],
        upload_read_timeout: int,
        output_file: Optional[typing.IO[Any]] = None,
    ) -> Optional[threading.Thread]:
        """Uploads intake file to running job, when intake type is localFile"""

        if intake_settings["type"] != IntakeAdapters.LOCAL_FILE:
            return None

        # There is source data to upload, so spin up a thread to handle
        # the upload concurrently and for thread safety issues, make
        # a copy of the REST client object
        _upload_client = cls._client.copy()
        job_csv_settings = job_response.get("jobSpec", {}).get("csvSettings", {})
        upload_thread = None

        def _get_file_size(fileobj: typing.BinaryIO) -> int:
            # To cover both files and filelike obj utilize .tell
            cur = fileobj.tell()
            fileobj.seek(0, os.SEEK_END)
            file_size = fileobj.tell()
            fileobj.seek(cur)

            return file_size

        # pylint: disable-next=unused-argument
        def _create_csv_chunk(
            header: Iterable[Any],
            reader: Iterator[Any],
            max_size: int,
            delimiter: str,
            encoding: str,
            quotechar: str,
        ) -> Tuple[io.StringIO, int]:
            chunk = io.StringIO()
            bytes_written = 0
            writer = csv.writer(chunk, delimiter=delimiter, quotechar=quotechar)
            writer.writerow(header)
            while bytes_written < max_size:
                try:
                    csv_chunk_content = next(reader)
                    written = writer.writerow(csv_chunk_content)
                    bytes_written += written
                except StopIteration:
                    break

            return chunk, bytes_written

        def _fileobj_to_csv_stream(fileobj: typing.BinaryIO, encoding: str) -> Iterator[Any]:
            stream = io.TextIOWrapper(fileobj, encoding=encoding)

            yield from csv.reader(stream)

            stream.close()

        def _upload_multipart(fileobj: typing.BinaryIO, base_upload_url: str) -> None:
            is_async = intake_settings.get("async", True)
            MAX_RETRY = 1 if is_async else 3
            MB_PER_CHUNK = 5
            CHUNK_MAX_SIZE = MB_PER_CHUNK * 1024 * 1024

            delimiter = job_csv_settings.get("delimiter", ",")
            encoding = job_csv_settings.get("encoding", "utf-8")
            quotechar = job_csv_settings.get("quotechar", '"')

            file_size = _get_file_size(fileobj)
            csv_stream = _fileobj_to_csv_stream(fileobj, encoding)

            # grab the header so it can be added on all parts
            header = next(csv_stream)

            part_number = 0
            bytes_written = 0
            while bytes_written <= file_size:
                part_upload_url = f"{base_upload_url}part/{part_number}"

                # Read the inputfile in chunks of CHUNK_MAX_SIZE
                # Then call put multiple times increasing the part_number each time
                chunk, chunk_bytes = _create_csv_chunk(
                    header, csv_stream, CHUNK_MAX_SIZE, delimiter, encoding, quotechar
                )
                bytes_written += chunk_bytes
                if chunk_bytes == 0:
                    break

                for attempts in range(MAX_RETRY):
                    try:
                        chunk.seek(0)
                        response = _upload_client.put(
                            url=part_upload_url,
                            data=chunk,
                            headers={"content-type": "text/csv"},
                            timeout=(_upload_client.connect_timeout, upload_read_timeout),
                        )

                        # Success! don't retry
                        if response.status_code == 202:
                            chunk.close()
                            break
                    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
                        attempts += 1
                        if attempts == MAX_RETRY:
                            raise

                part_number += 1

            # finalize the upload to indicate no more data is arriving
            _upload_client.post(url=f"{base_upload_url}finalizeMultipart")

        def _uploader() -> None:
            upload_url = job_response["links"]["csvUpload"]

            if "file_path" in intake_file:
                fileobj = open(  # pylint: disable=consider-using-with
                    intake_file["file_path"], "rb"
                )
            else:
                fileobj = intake_file["filelike"]
                fileobj.seek(0)

            try:
                if intake_settings.get("multipart"):
                    _upload_multipart(fileobj, upload_url)

                else:
                    _upload_client.put(
                        url=upload_url,
                        data=fileobj,
                        headers={"content-type": "text/csv"},
                        timeout=(_upload_client.connect_timeout, upload_read_timeout),
                    )
            finally:
                if hasattr(fileobj, "close"):
                    fileobj.close()

        if output_file is not None:
            # If output_file is specified, we upload and download
            # concurrently

            upload_thread = threading.Thread(target=_uploader)
            upload_thread.setDaemon(True)
            upload_thread.start()

        else:
            # Otherwise, upload is synchronous

            _uploader()
        return upload_thread

    @classmethod
    def _job_path(cls, project_id: str, job_id: str) -> str:
        raise NotImplementedError

    def _delete(self, ignore_404_errors: bool = False) -> None:
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        status = self._get_status()

        prediction_job_id = status["links"]["self"].split("/")[-2]
        try:
            self._client.delete(self._job_path(project_id="", job_id=prediction_job_id))
        except errors.ClientError as exc:
            if exc.status_code == 404 and ignore_404_errors:
                return
            raise

    def _get_status(self) -> Any:
        """Get status of batch job"""
        batch_job = super().get(project_id="", job_id=self.id)
        batch_job.id = self.id

        return batch_job._safe_data

    def _download(
        self, fileobj: typing.IO[Any], timeout: int = 120, read_timeout: int = 660
    ) -> None:
        """Downloads the CSV result of a batch job"""

        status = self._wait_for_download(timeout=timeout)
        download_iter = self._client.get(
            status["links"]["download"],
            stream=True,
            timeout=read_timeout,
        ).iter_content(chunk_size=8192)

        for chunk in download_iter:
            if chunk:
                fileobj.write(chunk)

        # Check if job was aborted during download (and the download is incomplete)
        status = self._get_status()
        if status["status"] in ("ABORTED", "FAILED"):
            raise RuntimeError("Job {} was aborted: {}".format(self.id, status["status_details"]))

    def _wait_for_download(self, timeout: int = 120) -> Any:
        """Waits for download to become available"""
        start = time.time()
        status = None
        while True:
            status = self._get_status()

            output_adapter_type = status["job_spec"].get("output_settings", {}).get("type")
            if output_adapter_type and not output_adapter_type == OutputAdapters.LOCAL_FILE:
                raise RuntimeError(
                    (
                        "You cannot download predictions from jobs that did not use local_file as "
                        "the output adapter. Job with ID {} had the output adapter defined as {}."
                    ).format(self.id, output_adapter_type)
                )

            if status["status"] in ("ABORTED", "FAILED"):
                raise RuntimeError(
                    "Job {} was aborted: {}".format(self.id, status["status_details"])
                )

            if "download" in status["links"]:
                break

            if timeout >= 0 and time.time() - start > timeout:  # pylint: disable=chained-comparison
                break

            time.sleep(1)

        if "download" not in status["links"]:
            # Ignore 404 errors here if the job never started - then we can't abort it
            self._delete(ignore_404_errors=True)
            raise RuntimeError(
                (
                    "Timed out waiting for download to become available for job ID {}. "
                    "Other jobs may be occupying the queue. Consider raising the timeout."
                ).format(self.id)
            )

        return status



# flake8: noqa

from .registered_model import (
    RegisteredModel,
    RegisteredModelListFilters,
    RegisteredModelVersionsListFilters,
)
from .registered_model_version import RegisteredModelVersion



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, List, Tuple

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject


class Cluster(APIObject):
    """Representation of a single cluster.

    Attributes
    ----------
    name: str
        Current cluster name
    percent: float
        Percent of data contained in the cluster. This value is reported after cluster insights
        are computed for the model.

    """

    _cluster_names = "projects/{project_id}/models/{model_id}/clusterNames/"

    _converter = t.Dict(
        {t.Key("name"): String(), t.Key("percent", optional=True): t.Float()}
    ).ignore_extra("*")

    def __init__(self, **kwargs: Any) -> None:
        self.name = kwargs.get("name")
        self.percent = kwargs.get("percent")

    def __repr__(self) -> str:
        return "Cluster(name={0.name}, percent={0.percent})".format(self)

    @classmethod
    def list(cls, project_id: str, model_id: str) -> List[Cluster]:
        """Retrieve a list of clusters in the model.

        Parameters
        ----------
        project_id: str
            ID of the project that the model is part of.
        model_id: str
            ID of the model.

        Returns
        -------
        List of clusters
        """
        path = cls._cluster_names.format(project_id=project_id, model_id=model_id)
        response = cls._client.get(path).json()
        return [cls.from_server_data(item) for item in response["clusters"]]

    @classmethod
    def update_multiple_names(
        cls, project_id: str, model_id: str, cluster_name_mappings: List[Tuple[str, str]]
    ) -> List[Cluster]:
        """Update many clusters at once based on list of name mappings.

        Parameters
        ----------
        project_id: str
            ID of the project that the model is part of.
        model_id: str
            ID of the model.
        cluster_name_mappings: List of tuples
            Cluster name mappings, consisting of current and previous names for each cluster.
            Example:

            .. code-block:: python

                cluster_name_mappings = [
                    ("current cluster name 1", "new cluster name 1"),
                    ("current cluster name 2", "new cluster name 2")]

        Returns
        -------
        List of clusters

        Raises
        ------
        datarobot.errors.ClientError
            Server rejected update of cluster names.
        ValueError
            Invalid cluster name mapping provided.
        """
        path = cls._cluster_names.format(project_id=project_id, model_id=model_id)
        mappings = [
            {"currentName": current_name, "newName": new_name}
            for current_name, new_name in cluster_name_mappings
        ]
        payload = {"clusterNameMappings": mappings}
        response = cls._client.patch(path, data=payload).json()
        return [cls.from_server_data(item) for item in response["clusters"]]

    @classmethod
    def update_name(
        cls, project_id: str, model_id: str, current_name: str, new_name: str
    ) -> List[Cluster]:
        """Change cluster name from current_name to new_name

        Parameters
        ----------
        project_id: str
           ID of the project that the model is part of.
        model_id: str
            ID of the model.
        current_name: str
            Current cluster name
        new_name: str
            New cluster name

        Returns
        -------
        List of Cluster
        """
        return cls.update_multiple_names(
            project_id=project_id,
            model_id=model_id,
            cluster_name_mappings=[(current_name, new_name)],
        )



# flake8: noqa
from ._version import __version__
from .client import Client
from .context import Context
from .enums import (
    AUTOPILOT_MODE,
    KeyValueCategory,
    KeyValueEntityType,
    KeyValueType,
    NETWORK_EGRESS_POLICY,
    PredictionEnvironmentPlatform,
    QUEUE_STATUS,
    SCORING_TYPE,
    SNAPSHOT_POLICY,
    TARGET_TYPE,
    VERBOSITY_LEVEL,
)
from .errors import AppPlatformError
from .helpers import *
from .models import (
    Application,
    AutomatedDocument,
    BatchMonitoringJob,
    BatchMonitoringJobDefinition,
    BatchPredictionJob,
    BatchPredictionJobDefinition,
    BlenderModel,
    Blueprint,
    BlueprintChart,
    BlueprintTaskDocument,
    CalendarFile,
    Cluster,
    ClusteringModel,
    ClusterInsight,
    CombinedModel,
    ComplianceDocTemplate,
    Connector,
    Credential,
    CustomInferenceModel,
    CustomModelTest,
    CustomModelVersion,
    CustomModelVersionDependencyBuild,
    CustomTask,
    CustomTaskVersion,
    DataDriver,
    DataEngineQueryGenerator,
    Dataset,
    DatasetDetails,
    DatasetFeature,
    DatasetFeatureHistogram,
    DatasetFeaturelist,
    DataSlice,
    DataSliceSizeInfo,
    DataSource,
    DataSourceParameters,
    DataStore,
    DatetimeModel,
    Deployment,
    ExecutionEnvironment,
    ExecutionEnvironmentVersion,
    ExternalConfusionChart,
    ExternalLiftChart,
    ExternalMulticlassLiftChart,
    ExternalResidualsChart,
    ExternalRocCurve,
    ExternalScores,
    Feature,
    FeatureAssociationFeaturelists,
    FeatureAssociationMatrix,
    FeatureAssociationMatrixDetails,
    FeatureHistogram,
    FeatureImpactJob,
    FeatureLineage,
    Featurelist,
    FrozenModel,
    genai,
    InteractionFeature,
    Job,
    KeyValue,
    Model,
    ModelBlueprintChart,
    ModelingFeature,
    ModelingFeaturelist,
    ModelJob,
    ModelRecommendation,
    PayoffMatrix,
    PredictionDataset,
    PredictionEnvironment,
    PredictionExplanations,
    PredictionExplanationsInitialization,
    Predictions,
    PredictionServer,
    PredictJob,
    PrimeFile,
    PrimeModel,
    Project,
    RatingTable,
    RatingTableModel,
    RegisteredModel,
    RegisteredModelVersion,
    registry,
    RelationshipsConfiguration,
    Ruleset,
    SecondaryDatasetConfigurations,
    SegmentationTask,
    SegmentInfo,
    ShapImpact,
    ShapMatrix,
    ShapMatrixJob,
    SharingAccess,
    SharingRole,
    TrainingPredictions,
    TrainingPredictionsJob,
    UseCase,
    UserBlueprint,
)

DR_TRACKABLE = True



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datarobot.models.genai.custom_model_validation import CustomModelValidation


class CustomModelLLMValidation(CustomModelValidation):
    """
    Validation record checking the ability of the deployment to serve
    as a custom model LLM.

    Attributes
    ----------
    prompt_column_name : str
        The column name the deployed model expect as the input.
    target_column_name : str
        The target name that the deployed model will output.
    deployment_id : str
        ID of the deployment.
    model_id : str
        ID of the underlying deployment model.
        Can be found from the API as Deployment.model["id"].
    validation_status : str
        Can be TESTING, FAILED, or PASSED. Only PASSED is allowed for use.
    deployment_access_data : dict, optional
        Data that will be used for accessing deployment prediction server.
        Only available for deployments that passed validation. Dict fields:
        - prediction_api_url - URL for deployment prediction server.
        - datarobot_key - first of 2 auth headers for the prediction server.
        - authorization_header - second of 2 auth headers for the prediction server.
        - input_type - Either JSON or CSV - the input type that the model expects.
        - model_type - Target type of the deployed custom model.
    tenant_id : str
        Creating user's tenant ID.
    error_message : Optional[str]
        Additional information for errored validation.
    deployment_name : Optional[str]
        The name of the deployment that is validated.
    user_name : Optional[str]
        The name of the user
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    prediction_timeout: int
        The timeout in seconds for the prediction API used in this custom model validation.
    """

    _path = "api/v2/genai/customModelLLMValidations"



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import datetime
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import dateutil
from mypy_extensions import TypedDict
import trafaret as t

from datarobot.enums import (
    RegisteredModelDeploymentSortKey,
    RegisteredModelSortDirection,
    RegisteredModelSortKey,
    RegisteredModelVersionSortKey,
)
from datarobot.mixins.browser_mixin import BrowserMixin
from datarobot.models.api_object import APIObject
from datarobot.models.model_registry.common import UserMetadata
from datarobot.models.model_registry.deployment import VersionAssociatedDeployment
from datarobot.models.model_registry.registered_model_version import RegisteredModelVersion
from datarobot.models.sharing import SharingRole
from datarobot.utils.pagination import unpaginate


class Target(TypedDict):
    name: str
    type: str


class RegisteredModelVersionsListFilters:
    """
    Filters for listing of registered model versions.

    Parameters
    ----------
    target_name: str or None
        Name of the target to filter by.
    target_type: str or None
        Type of the target to filter by.
    compatible_with_leaderboard_model_id: str or None.
        If specified, limit results to versions (model packages) of the Leaderboard model with the specified ID.
    compatible_with_model_package_id: str or None.
        Returns versions compatible with the given model package (version) ID. If used, it will only return versions
        that match `target.name`, `target.type`, `target.classNames` (for classification models),
        `modelKind.isTimeSeries` and `modelKind.isMultiseries` for the specified model package (version).
    for_challenger: bool or None
        Can be used with compatibleWithModelPackageId to request similar versions that can be used as challenger
        models; for external model packages (versions), instead of returning similar external model packages (versions),
        similar DataRobot and Custom model packages (versions) will be retrieved.
    prediction_threshold: float or None
        Return versions with the specified prediction threshold used for binary classification models.
    imported: bool or None
        If specified, return either imported (true) or non-imported (false) versions (model packages).
    prediction_environment_id: str or None
        Can be used to filter versions (model packages) by what is supported by the prediction environment
    model_kind: str or None
        Can be used to filter versions (model packages) by model kind.
    build_status: str or None
        If specified, filter versions by the build status.
    """

    def __init__(
        self,
        target_name: Optional[str] = None,
        target_type: Optional[str] = None,
        compatible_with_leaderboard_model_id: Optional[str] = None,
        compatible_with_model_package_id: Optional[str] = None,
        for_challenger: Optional[bool] = None,
        prediction_threshold: Optional[float] = None,
        imported: Optional[bool] = None,
        prediction_environment_id: Optional[str] = None,
        model_kind: Optional[str] = None,
        build_status: Optional[str] = None,
        use_case_id: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ):
        self.target_name = target_name
        self.target_type = target_type
        self.compatible_with_leaderboard_model_id = compatible_with_leaderboard_model_id
        self.compatible_with_model_package_id = compatible_with_model_package_id
        self.for_challenger = for_challenger
        self.prediction_threshold = prediction_threshold
        self.imported = imported
        self.prediction_environment_id = prediction_environment_id
        self.model_kind = model_kind
        self.build_status = build_status
        self.use_case_id = use_case_id
        self.tags = tags

    def _construct_query_args(self) -> Dict[str, Any]:
        """
        Construct query args for the list endpoint.
        Returns
        -------
        query_args : Dict[str, Any]
            Dictionary of query args.
        """
        args: Dict[str, Any] = {}
        if self.target_name is not None:
            args["targetName"] = self.target_name
        if self.target_type is not None:
            args["targetType"] = self.target_type
        if self.compatible_with_leaderboard_model_id is not None:
            args["compatibleWithLeaderboardModelId"] = self.compatible_with_leaderboard_model_id
        if self.compatible_with_model_package_id is not None:
            args["compatibleWithModelPackageId"] = self.compatible_with_model_package_id
        if self.for_challenger is not None:
            args["forChallenger"] = self.for_challenger
        if self.prediction_threshold is not None:
            args["predictionThreshold"] = self.prediction_threshold
        if self.imported is not None:
            args["imported"] = self.imported
        if self.prediction_environment_id is not None:
            args["predictionEnvironmentId"] = self.prediction_environment_id
        if self.model_kind is not None:
            args["modelKind"] = self.model_kind
        if self.build_status is not None:
            args["buildStatus"] = self.build_status
        if self.use_case_id is not None:
            args["useCaseId"] = self.use_case_id
        if self.tags is not None:
            args["tags"] = self.tags
        return args


class RegisteredModelListFilters:
    """
    Filters for listing registered models.

    Parameters
    ----------
    created_at_start : datetime.datetime
        Registered models created on or after this timestamp.
    created_at_end : datetime.datetime
        Registered models created before this timestamp. Defaults to the current time.
    modified_at_start : datetime.datetime
        Registered models modified on or after this timestamp.
    modified_at_end : datetime.datetime
        Registered models modified before this timestamp. Defaults to the current time.
    target_name : str
        Name of the target to filter by.
    target_type : str
        Type of the target to filter by.
    created_by : str
        Email of the user that created registered model to filter by.
    compatible_with_leaderboard_model_id : str
        If specified, limit results to registered models containing versions (model packages)
        for the leaderboard model with the specified ID.
    compatible_with_model_package_id : str
        Return registered models that have versions (model packages) compatible with given model package (version) ID.
        If used, will only return registered models which have versions that match `target.name`, `target.type`,
        `target.classNames` (for classification models), `modelKind.isTimeSeries`, and `modelKind.isMultiseries`
        of the specified model package (version).
    for_challenger : bool
        Can be used with compatibleWithModelPackageId to request similar registered models that contain
        versions (model packages) that can be used as challenger models; for external model packages (versions),
        instead of returning similar external model packages (versions), similar DataRobot and Custom model packages
        will be retrieved.
    prediction_threshold : float
        If specified, return any registered models containing one or more versions matching the prediction
        threshold used for binary classification models.
    imported : bool
        If specified, return any registered models that contain either imported (true) or non-imported (false)
        versions (model packages).
    prediction_environment_id : str
        Can be used to filter registered models by what is supported by the prediction environment.
    model_kind : str
        Return models that contain versions matching a specific format.
    build_status : str
        If specified, only return models that have versions with specified build status.
    """

    def __init__(
        self,
        created_at_start: Optional[datetime.datetime] = None,
        created_at_end: Optional[datetime.datetime] = None,
        modified_at_start: Optional[datetime.datetime] = None,
        modified_at_end: Optional[datetime.datetime] = None,
        target_name: Optional[str] = None,
        target_type: Optional[str] = None,
        created_by: Optional[str] = None,
        compatible_with_leaderboard_model_id: Optional[str] = None,
        compatible_with_model_package_id: Optional[str] = None,
        for_challenger: Optional[bool] = None,
        prediction_threshold: Optional[float] = None,
        imported: Optional[bool] = None,
        prediction_environment_id: Optional[str] = None,
        model_kind: Optional[str] = None,
        build_status: Optional[str] = None,
    ):
        self.created_at_start = created_at_start
        self.created_at_end = created_at_end
        self.modified_at_start = modified_at_start
        self.modified_at_end = modified_at_end
        self.target_name = target_name
        self.target_type = target_type
        self.created_by = created_by
        self.compatible_with_leaderboard_model_id = compatible_with_leaderboard_model_id
        self.compatible_with_model_package_id = compatible_with_model_package_id
        self.for_challenger = for_challenger
        self.prediction_threshold = prediction_threshold
        self.imported = imported
        self.prediction_environment_id = prediction_environment_id
        self.model_kind = model_kind
        self.build_status = build_status

    def _construct_query_args(self) -> Dict[str, Any]:
        """
        Construct query args for the list endpoint.
        Returns
        -------
        query_args : Dict[str, Any]
            Dictionary of query args.
        """
        query_args: Dict[str, Any] = {}
        if self.created_at_start:
            query_args["createdAtStartTs"] = self.created_at_start.isoformat()
        if self.created_at_end:
            query_args["createdAtEndTs"] = self.created_at_end.isoformat()
        if self.modified_at_start:
            query_args["modifiedAtStartTs"] = self.modified_at_start.isoformat()
        if self.modified_at_end:
            query_args["modifiedAtEndTs"] = self.modified_at_end.isoformat()
        if self.target_name:
            query_args["targetName"] = self.target_name
        if self.target_type:
            query_args["targetType"] = self.target_type
        if self.created_by:
            query_args["createdBy"] = self.created_by
        if self.compatible_with_leaderboard_model_id:
            query_args[
                "compatibleWithLeaderboardModelId"
            ] = self.compatible_with_leaderboard_model_id
        if self.compatible_with_model_package_id:
            query_args["compatibleWithModelPackageId"] = self.compatible_with_model_package_id
        if self.for_challenger:
            query_args["forChallenger"] = self.for_challenger
        if self.prediction_threshold:
            query_args["predictionThreshold"] = self.prediction_threshold
        if self.imported is not None:
            query_args["imported"] = self.imported
        if self.prediction_environment_id:
            query_args["predictionEnvironmentId"] = self.prediction_environment_id
        if self.model_kind:
            query_args["modelKind"] = self.model_kind
        if self.build_status:
            query_args["buildStatus"] = self.build_status
        return query_args


TRegisteredModel = TypeVar("TRegisteredModel", bound="RegisteredModel")


class RegisteredModel(APIObject, BrowserMixin):
    """A registered model is a logical grouping of model packages (versions) that are related to each other.

    Attributes
    ----------
    id : str
        The ID of the registered model.
    name : str
        The name of the registered model.
    description : str
        The description of the registered model.
    created_at : str
        The creation time of the registered model.
    modified_at : str
        The last modification time for the registered model.
    modified_by : datarobot.models.model_registry.common.UserMetadata
        Information on the user who last modified the registered model.
    target : Target
        Information on the target variable.
    created_by : datarobot.models.model_registry.common.UserMetadata
        Information on the creator of the registered model.
    last_version_num : int
        The latest version number associated to this registered model.
    is_archived : bool
         Determines whether the registered model is archived.
    """

    _path = "registeredModels/"

    _user_metadata = t.Dict().allow_extra("*")
    _target = t.Dict().allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("name"): t.String(),
            t.Key("description", optional=True): t.Or(t.String(allow_blank=True), t.Null),
            t.Key("created_at"): t.String() >> dateutil.parser.parse,
            t.Key("modified_at"): t.String() >> dateutil.parser.parse,
            t.Key("modified_by", optional=True): t.Or(_user_metadata, t.Null),
            t.Key("target"): _target,
            t.Key("created_by"): _user_metadata,
            t.Key("last_version_num"): t.Int(),
            t.Key("is_archived"): t.Bool(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        description: Optional[str],
        created_at: str,
        modified_at: str,
        target: Target,
        created_by: UserMetadata,
        last_version_num: int,
        is_archived: bool,
        modified_by: Optional[UserMetadata] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.created_at = created_at
        self.modified_at = modified_at
        self.modified_by = modified_by
        self.target = target
        self.created_by = created_by
        self.last_version_num = last_version_num
        self.is_archived = is_archived

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    @classmethod
    def get(cls: Type[TRegisteredModel], registered_model_id: str) -> TRegisteredModel:
        """
        Get a registered model by ID.

        Parameters
        ----------
        registered_model_id : str
            ID of the registered model to retrieve

        Returns
        -------
        registered_model : RegisteredModel
            Registered Model Object

        Examples
        --------
        .. code-block:: python

            from datarobot import RegisteredModel
            registered_model = RegisteredModel.get(registered_model_id='5c939e08962d741e34f609f0')
            registered_model.id
            >>>'5c939e08962d741e34f609f0'
            registered_model.name
            >>>'My Registered Model'
        """
        path = f"{cls._path}{registered_model_id}/"
        return cls.from_location(path)

    @classmethod
    def list(
        cls: Type[TRegisteredModel],
        limit: Optional[int] = 100,
        offset: Optional[int] = None,
        sort_key: Optional[RegisteredModelSortKey] = None,
        sort_direction: Optional[RegisteredModelSortDirection] = None,
        search: Optional[str] = None,
        filters: Optional[RegisteredModelListFilters] = None,
    ) -> List[TRegisteredModel]:
        """
        List all registered models a user can view.

        Parameters
        ----------
        limit : int, optional
            Maximum number of registered models to return
        offset : int, optional
            Number of registered models to skip before returning results
        sort_key : RegisteredModelSortKey, optional
            Key to order result by
        sort_direction : RegisteredModelSortDirection, optional
            Sort direction
        search : str, optional
            A term to search for in registered model name, description, or target name
        filters : RegisteredModelListFilters, optional
            An object containing all filters that you'd like to apply to the
            resulting list of registered models.
        Returns
        -------
        registered_models : List[RegisteredModel]
            A list of registered models user can view.

        Examples
        --------
        .. code-block:: python

            from datarobot import RegisteredModel
            registered_models = RegisteredModel.list()
            >>> [RegisteredModel('My Registered Model'), RegisteredModel('My Other Registered Model')]

        .. code-block:: python

            from datarobot import RegisteredModel
            from datarobot.models.model_registry import RegisteredModelListFilters
            from datarobot.enums import RegisteredModelSortKey, RegisteredModelSortDirection
            filters = RegisteredModelListFilters(target_type='Regression')
            registered_models = RegisteredModel.list(
                filters=filters,
                sort_key=RegisteredModelSortKey.NAME.value,
                sort_direction=RegisteredModelSortDirection.DESC.value
                search='other')
            >>> [RegisteredModel('My Other Registered Model')]

        """
        if filters is None:
            filters = RegisteredModelListFilters()
        params: Dict[str, Union[int, str]] = {}
        if sort_key:
            params["sortKey"] = sort_key.value
        if sort_direction:
            params["sortDirection"] = sort_direction.value
        if search:
            params["search"] = search
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        params.update(filters._construct_query_args())
        data = unpaginate(cls._path, params, cls._client)
        return [cls.from_server_data(data_point) for data_point in data]

    @classmethod
    def archive(cls, registered_model_id: str) -> None:
        """
        Permanently archive a registered model and all of its versions.

        Parameters
        ----------
        registered_model_id : str
            ID of the registered model to be archived

        Returns
        -------

        """
        url = f"{cls._path}{registered_model_id}/"
        cls._client.delete(url)

    @classmethod
    def update(
        cls: Type[TRegisteredModel], registered_model_id: str, name: str
    ) -> TRegisteredModel:
        """
        Update the name of a registered model.

        Parameters
        ----------
        registered_model_id : str
            ID of the registered model to be updated
        name : str
            New name for the registered model

        Returns
        -------
        registered_model : RegisteredModel
            Updated registered model object

        """
        url = f"{cls._path}{registered_model_id}/"
        data = cls._client.patch(url, data={"name": name})
        return cls.from_server_data(data.json())

    def get_shared_roles(
        self,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        id: Optional[str] = None,
    ) -> List[SharingRole]:
        """
        Retrieve access control information for this registered model.

        Parameters
        ----------
        offset : Optional[int]
            The number of records to skip over. Optional. Default is 0.
        limit: Optional[int]
            The number of records to return. Optional. Default is 100.
        id: Optional[str]
            Return the access control information for a user with this user ID. Optional.
        """
        params = {"offset": offset, "limit": limit, "id": id}
        path = f"{self._path}{self.id}/sharedRoles/"
        r_data = unpaginate(path, params, self._client)
        return [SharingRole.from_server_data(data) for data in r_data]

    def share(self, roles: List[SharingRole]) -> None:
        """
        Share this registered model or remove access from one or more user(s).

        Parameters
        ----------
        roles : List[SharingRole]
            A list of :class:`SharingRole <datarobot.models.sharing.SharingRole>` instances, each of which
            references a user and a role to be assigned.

        Examples
        --------
        .. code-block:: python

           >>> from datarobot import RegisteredModel, SharingRole
           >>> from datarobot.enums import SHARING_ROLE, SHARING_RECIPIENT_TYPE
           >>> registered_model = RegisteredModel.get('5c939e08962d741e34f609f0')
           >>> sharing_role = SharingRole(
           ...    role=SHARING_ROLE.CONSUMER,
           ...    recipient_type=SHARING_RECIPIENT_TYPE.USER,
           ...    id='5c939e08962d741e34f609f0',
           ...    can_share=True,
           ...    )
           >>> registered_model.share(roles=[sharing_role])

        """
        path = f"{self._path}{self.id}/sharedRoles/"
        formatted_roles = [role.collect_payload() for role in roles]
        payload = {"roles": formatted_roles, "operation": "updateRoles"}
        self._client.patch(path, data=payload)

    def get_version(self, version_id: str) -> RegisteredModelVersion:
        """
        Retrieve a registered model version.

        Parameters
        ----------
        version_id : str
            The ID of the registered model version to retrieve.

        Returns
        -------
        registered_model_version : RegisteredModelVersion
            A registered model version object.

        Examples
        --------
        .. code-block:: python

            from datarobot import RegisteredModel
            registered_model = RegisteredModel.get('5c939e08962d741e34f609f0')
            registered_model_version = registered_model.get_version('5c939e08962d741e34f609f0')
            >>> RegisteredModelVersion('My Registered Model Version')

        """
        path = f"{self._path}{self.id}/versions/{version_id}/"
        data = self._client.get(path).json()
        return RegisteredModelVersion.from_server_data(data)

    def list_versions(
        self,
        filters: Optional[RegisteredModelVersionsListFilters] = None,
        search: Optional[str] = None,
        sort_key: Optional[RegisteredModelVersionSortKey] = None,
        sort_direction: Optional[RegisteredModelSortDirection] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[RegisteredModelVersion]:
        """
        Retrieve a list of registered model versions.

        Parameters
        ----------
        filters : Optional[RegisteredModelVersionsListFilters]
            A RegisteredModelVersionsListFilters instance used to filter the list of registered model versions returned.
        search : Optional[str]
            A search string used to filter the list of registered model versions returned.
        sort_key : Optional[RegisteredModelVersionSortKey]
            The key to use to sort the list of registered model versions returned.
        sort_direction : Optional[RegisteredModelSortDirection]
            The direction to use to sort the list of registered model versions returned.
        limit : Optional[int]
            The maximum number of registered model versions to return. Default is 100.
        offset : Optional[int]
            The number of registered model versions to skip over. Default is 0.

        Returns
        -------
        registered_model_versions : List[RegisteredModelVersion]
            A list of registered model version objects.

        Examples
        --------
        .. code-block:: python

            from datarobot import RegisteredModel
            from datarobot.models.model_registry import RegisteredModelVersionsListFilters
            from datarobot.enums import RegisteredModelSortKey, RegisteredModelSortDirection
            registered_model = RegisteredModel.get('5c939e08962d741e34f609f0')
            filters = RegisteredModelVersionsListFilters(tags=['tag1', 'tag2'])
            registered_model_versions = registered_model.list_versions(filters=filters)
            >>> [RegisteredModelVersion('My Registered Model Version')]
        """
        if filters is None:
            filters = RegisteredModelVersionsListFilters()
        params: Dict[str, Any] = {}
        if sort_key:
            params["sortKey"] = sort_key.value
        if sort_direction:
            params["sortDirection"] = sort_direction.value
        if search:
            params["search"] = search
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        url = f"{self._path}{self.id}/versions/"
        params.update(filters._construct_query_args())
        data = unpaginate(url, params, self._client)
        return [RegisteredModelVersion.from_server_data(data_point) for data_point in data]

    def list_associated_deployments(
        self,
        search: Optional[str] = None,
        sort_key: Optional[RegisteredModelDeploymentSortKey] = None,
        sort_direction: Optional[RegisteredModelSortDirection] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[VersionAssociatedDeployment]:
        """
        Retrieve a list of deployments associated with this registered model.

        Parameters
        ----------
        search : Optional[str]
        sort_key : Optional[RegisteredModelDeploymentSortKey]
        sort_direction : Optional[RegisteredModelSortDirection]
        limit : Optional[int]
        offset : Optional[int]

        Returns
        -------
        deployments : List[VersionAssociatedDeployment]
            A list of deployments associated with this registered model.

        """
        params: Dict[str, Any] = {}
        if sort_key:
            params["sortKey"] = sort_key.value
        if sort_direction:
            params["sortDirection"] = sort_direction.value
        if search:
            params["search"] = search
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        url = f"{self._path}{self.id}/deployments/"
        data = unpaginate(url, params, self._client)
        return [VersionAssociatedDeployment.from_server_data(data_point) for data_point in data]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
This module contains compatibility fixes to allow usage of both 1.x and 2.x Trafaret versions
"""
try:
    from trafaret import ToInt as Int  # pylint: disable=unused-import
except ImportError:
    from trafaret import Int  # noqa pylint: disable=unused-import


try:
    from trafaret import AnyString as String  # pylint: disable=unused-import
except ImportError:
    from trafaret import String  # noqa pylint: disable=unused-import



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, cast, Dict, List, Optional, Union

import trafaret as t

from datarobot.enums import ENTITY_TYPES, INSIGHTS_SOURCES
from datarobot.insights.base import BaseInsight


class ShapImpact(BaseInsight):
    """Class for SHAP Impact calculations. Use the standard methods of BaseInsight to compute and
    retrieve SHAP matrices:
    * compute: submit a request to compute a SHAP impact, and return immediately
    * create: submit a request to compute a SHAP impact, and wait for it to finish
    * list: retrieve all ShapImpact results for a model, possibly on multiple datasets or data slices.
    """

    INSIGHT_NAME = "shapImpact"
    INSIGHT_DATA = {
        "shap_impacts": t.Or(
            t.List(t.List(t.Or(t.Int(), t.Float()))),
            t.Dict(
                {
                    t.Key("featureName"): t.String(),
                    t.Key("impactNormalized"): t.Float(),
                    t.Key("impactUnnormalized"): t.Float(),
                }
            ),
        ),
        "base_value": t.List(t.Float()),
        "capping": Optional[
            t.Or(
                t.Null(),
                t.Dict(
                    {
                        t.Key("right"): t.Or(t.String(), t.Float(), t.Null()),  # noqa: F821
                        t.Key("left"): t.Or(t.String(), t.Float(), t.Null()),  # noqa: F821
                    }
                ),
            )
        ],
        "link": t.Or(t.String(), t.Null()),
        "row_count": t.Int(),
    }

    @classmethod
    def _get_payload(
        cls,
        entity_id: str,
        source: str = INSIGHTS_SOURCES.TRAINING,
        data_slice_id: Optional[str] = None,
        external_dataset_id: Optional[str] = None,
        entity_type: Optional[ENTITY_TYPES] = None,
        row_count: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Union[str, int]]:
        """Construct a payload for a compute request"""
        payload = super()._get_payload(
            entity_id=entity_id,
            source=source,
            data_slice_id=data_slice_id,
            external_dataset_id=external_dataset_id,
            entity_type=entity_type,
            **kwargs,
        )

        if row_count is None:
            raise ValueError("row_count is required")
        payload["rowCount"] = row_count

        return payload

    @property
    def shap_impacts(self) -> List[List[Any]]:
        """SHAP impact values

        Returns
        -------
        shap impacts
            A list of the SHAP impact values
        """
        return cast(List[List[Any]], self.data["shap_impacts"])

    @property
    def base_value(self) -> List[float]:
        """A list of base prediction values"""
        return cast(List[float], self.data["base_value"])

    @property
    def capping(self) -> Optional[Dict[str, Any]]:
        """Capping for the models in the blender"""
        return cast(Optional[Dict[str, Any]], self.data.get("capping"))

    @property
    def link(self) -> Optional[str]:
        """Shared link function of the models in the blender"""
        return cast(Optional[str], self.data.get("link"))

    @property
    def row_count(self) -> int:
        """Number of SHAP impact rows"""
        return cast(int, self.data["row_count"])



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import inspect
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from datarobot.context import Context

TRACE_SEPARATOR = "->"

STACK_TRACE_HEADER = "X-DataRobot-Api-Consumer-Trace"
STACK_TRACE_MAX_CHARACTERS = 1000


def trackable(is_trackable: bool = True) -> Callable:  # type: ignore[type-arg]
    def wrapper(fn: Callable) -> Callable:  # type: ignore[type-arg]
        setattr(fn, "DR_TRACKABLE", is_trackable)
        return fn

    return wrapper


def is_trackable(obj: Any) -> bool:
    """
    Determine if a class, instance, or function is trackable for DataRobot usage tracking.
    """
    return not hasattr(obj, "DR_TRACKABLE") or getattr(obj, "DR_TRACKABLE")


def check_parent_package_is_trackable(
    package_name: str, trackable_packages: Dict[str, bool]
) -> Tuple[Dict[str, bool], bool]:
    """
    Check a package's parent packages to see if any of them are listed as DR_TRACKABLE = True.
    DataRobot assumes that a package being trackable is inherited unless explicitly set in the package.

    Returns
    -------
    package_name : str
        Return the initial package name if a parent is listed as trackable.
    """
    package = sys.modules.get(package_name)
    package_update = {}
    while package:
        current_package_name = package.__name__
        # Assume a package isn't trackable until we find a parent that is
        package_update.update({current_package_name: False})
        if getattr(package, "DR_TRACKABLE", False) or trackable_packages.get(package.__name__):
            return {k: True for k in package_update}, True
        new_package_name = current_package_name.rpartition(".")[0]
        package = sys.modules.get(new_package_name, None)
    return package_update, False


def format_stack_frame(
    package: str,
    func: str,
    current_frame: Any,
    obj: Optional[Any] = None,
    cls: Optional[Type] = None,  # type: ignore[type-arg]
) -> Optional[str]:
    """
    Build an absolute stack trace string for a single stack frame.

    Returns
    -------
    stack_trace_string : str
        The resulting string.
        Example: `"datarobot.models.project.Project.get"`

    """
    stack_string = f"{package}"
    func_callable = current_frame.f_globals.get(current_frame.f_code.co_name)
    if obj:
        if not is_trackable(obj):
            return None
        stack_string = stack_string + f".{obj.__class__.__name__}"
    if cls:
        if not is_trackable(cls):
            return None
        stack_string = stack_string + f".{cls.__name__}"
    if func_callable and not is_trackable(func_callable):
        return None
    stack_string = stack_string + f".{func}"
    return stack_string


def get_stack_trace(stack_trace: Optional[List[inspect.FrameInfo]] = None) -> str:
    """
    Read a stack trace and generate a stack trace string to be sent to DataRobot for usage tracking.

    Returns
    -------
    stack_trace_string : str
        The complete stack trace for all frames that are within the `datarobot` package,
        ignoring all customer and 3rd party package code.
        Example: `"datarobot.models.project.Project.get->datarobot.models.api_object.Project.from_location"`
    """
    dr_trace = []
    if not stack_trace:
        stack_trace = inspect.stack()[1:]
    trackable_packages: Dict[str, bool] = {}
    for stack_frame in stack_trace:
        package_name = stack_frame.frame.f_globals.get("__package__")
        package = sys.modules.get(package_name, "")  # type: ignore[arg-type]
        if hasattr(package, "DR_TRACKABLE") and not getattr(package, "DR_TRACKABLE"):
            # if the package is explicitly listed as not trackable, immediately break.
            break
        if package_name not in trackable_packages:
            # otherwise, check if any of the parent packages are trackable
            package_update, package_ok = check_parent_package_is_trackable(
                package_name, trackable_packages  # type: ignore[arg-type]
            )
            trackable_packages.update(package_update)
            if not package_ok:
                break
        obj = stack_frame.frame.f_locals.get("self")
        if obj and "RESTClientObject" in obj.__class__.__name__:
            continue
        if not stack_frame.function.startswith("_"):
            trace_string = format_stack_frame(
                package=stack_frame.frame.f_globals.get("__name__"),  # type: ignore[arg-type]
                obj=obj,
                cls=stack_frame.frame.f_locals.get("cls"),
                func=stack_frame.function,
                current_frame=stack_frame.frame,
            )
            if trace_string:
                dr_trace.append(trace_string)
                continue
            else:
                break
    if Context.trace_context:
        dr_trace.append(Context.trace_context)
    return TRACE_SEPARATOR.join(reversed([trace_str for trace_str in dr_trace if trace_str]))[
        :STACK_TRACE_MAX_CHARACTERS
    ]
